"""
This file houses a series of functions that load, convert, and save vulnerability data with Parquet files.
"""
import argparse
import logging
import os
import pathlib
from math import ceil
from typing import Tuple, List

import numba as nb
import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq

from .common import Vulnerability

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


vulnerability_filename = 'vulnerability.bin'
vulnerability_dataset = "vulnerability_dataset"
vulnerability_parquet_filename = 'part_{}.parquet'

areaperil_int = np.dtype(os.environ.get('AREAPERIL_TYPE', 'u4'))
oasis_float = np.dtype(os.environ.get('OASIS_FLOAT', 'f4'))

vulnerability_bloc_size = int(os.environ.get('VULNERABILITY_BLOC_SIZE', 1e9))  # Bytes


@nb.njit(cache=True)
def get_vuln_info(vulns_bin) -> Tuple[int, int, int]:
    """
    Extracts meta data from the vulnerability data.

    Args:
        vulns_bin: (List[Vulnerability]) vulnerability data from the file

    Returns: (Tuple[int, int, int]) number of vulnerability IDs, number of intensity bins, number of damage bins
    """
    vulnerability_ids_set = set()
    num_intensity_bins = 0
    num_damage_bins = 0

    for vuln_i in range(vulns_bin.shape[0]):
        vuln = vulns_bin[vuln_i]
        vulnerability_ids_set.add(vuln['vulnerability_id'])
        if num_intensity_bins < vuln['intensity_bin_id']:
            num_intensity_bins = vuln['intensity_bin_id']

        if num_damage_bins < vuln['damage_bin_id']:
            num_damage_bins = vuln['damage_bin_id']

    return len(vulnerability_ids_set), num_intensity_bins, num_damage_bins


@nb.njit()
def get_array(vulns_bin, num_intensity_bins, num_damage_bins, max_vulnerability_id_size) -> Tuple[List[int], List[List[List[float]]]]:
    """
    Flattens the vulnerability data into a one-dimensional array for each vulnerability ID.

    NOTE => Numba: cannot cache generator for the moment to work properlly, data on the same vulnerability_id must
            all be in one block

    Args:
        vulns_bin: (List[Vulnerability]) vulnerability data from file
        num_intensity_bins: (int) the number of intensity bins in the data
        num_damage_bins: (int) the number of damage bins in the data
        max_vulnerability_id_size: (int) the size of the vulnerability_ids array inside the function, this will
                                         be the size of the vulnerability IDs array that is returned

    Returns: (Tuple[List[int], List[List[List[float]]]]) array of vulnerability IDs, vulnerability data
    """
    vulnerability_ids = np.empty(max_vulnerability_id_size, dtype=np.int32)
    vuln_array = np.zeros((vulnerability_ids.shape[0], num_damage_bins, num_intensity_bins), dtype=oasis_float)

    cursor = 0
    vulnerability_id_index = 0
    vulnerability_id = -1

    while cursor < vulns_bin.shape[0]:
        vuln: Vulnerability = vulns_bin[cursor]
        if vuln['vulnerability_id'] != vulnerability_id:
            if vulnerability_id_index == max_vulnerability_id_size:
                yield vulnerability_ids, vuln_array
                vuln_array.fill(0)
                vulnerability_id_index = 0
            vulnerability_id = vulns_bin[cursor]['vulnerability_id']
            vulnerability_ids[vulnerability_id_index] = vulnerability_id
            cur_vuln_array = vuln_array[vulnerability_id_index]
            vulnerability_id_index += 1

        cur_vuln_array[vuln['damage_bin_id'] - 1, vuln['intensity_bin_id'] - 1] = vuln['probability']
        cursor += 1
    if vulnerability_id_index:
        yield vulnerability_ids[:vulnerability_id_index], vuln_array[:vulnerability_id_index]


def iter_table(vulns_bin, num_intensity_bins, num_damage_bins, info, max_vulnerability_id_size):
    """
    Loops through the vulnerability data, converting it into one-dimensional arrays, in-turn converting this into
    PyArrow arrays and then converting into Tables.

    Args:
        vulns_bin: (List[Vulnerability]) vulnerability data from file
        num_intensity_bins: (int) the number of intensity bins in the data
        num_damage_bins: (int) the number of damage bins in the data
        info: (dict) meta data around the vulnerability data. It has to store:
                                                                num_vulnerability_id => number of vulnerability IDs
                                                                num_intensity_bins => number of intensity bins
                                                                num_damage_bins => number of damage bins
        max_vulnerability_id_size: (int) the size of the vulnerability_ids array

    Returns:
    """
    for vulnerability_ids, vuln_array in get_array(vulns_bin, num_intensity_bins, num_damage_bins, max_vulnerability_id_size):
        arr_vulnerability_ids = pa.array(vulnerability_ids)
        arr_vulnerability = pa.FixedSizeListArray.from_arrays(vuln_array.ravel(), num_intensity_bins * num_damage_bins)
        vulnerability_table = pa.Table.from_arrays([arr_vulnerability_ids, arr_vulnerability], names=[
                                                   'vulnerability_id', 'vuln_array'], metadata=info)
        yield vulnerability_table


def vulnerability_to_parquet(run_dir) -> None:
    """
    Converts the vulnerability data to parquet file and saves it.

    Args:
        run_dir: (str) the directory of the data where this process is to take place

    Returns: None
    """
    logger.debug(f'retrieving vulnerability info from {os.path.join(run_dir, vulnerability_filename)}')
    vulns_bin = np.memmap(os.path.join(run_dir, vulnerability_filename), dtype=Vulnerability, offset=4, mode='r')
    num_vulnerability_id, num_intensity_bins, num_damage_bins = get_vuln_info(vulns_bin)

    info = {"num_vulnerability_id": str(num_vulnerability_id),
            "num_intensity_bins": str(num_intensity_bins),
            "num_damage_bins": str(num_damage_bins),
            }

    logger.debug(f'{info}')

    dataset_path = pathlib.Path(os.path.join(run_dir, vulnerability_dataset))
    dataset_path.mkdir(exist_ok=True)
    for filepath in dataset_path.glob(vulnerability_parquet_filename.format('*')):
        os.remove(filepath)

    max_vulnerability_id_size = vulnerability_bloc_size // (num_intensity_bins * num_damage_bins * oasis_float.itemsize)
    num_step = ceil(num_vulnerability_id / max_vulnerability_id_size)
    for i, vuln_table in enumerate(iter_table(vulns_bin, num_intensity_bins, num_damage_bins, info, max_vulnerability_id_size)):
        logger.debug(f"step {i + 1}/{num_step}")
        pq.write_table(vuln_table, os.path.join(dataset_path, vulnerability_parquet_filename.format(i)))


parser = argparse.ArgumentParser()
parser.add_argument('-r', '--run-dir', help='path to the directory containing vulnerability.bin', default='static')
parser.add_argument('-v', '--logging-level', help='logging level (debug:10, info:20, warning:30, error:40, critical:50)',
                    default=30, type=int)


def main():
    kwargs = vars(parser.parse_args())

    # add handler to fm logger
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    vulnerability_to_parquet(**kwargs)


if __name__ == "__main__":
    main()



"""
This file defines the data types that are loaded from the data files.

"""
import numpy as np
import numba as nb

from oasislmf.pytools.getmodel.common import oasis_float, areaperil_int

# gul stream type
# probably need to set this dynamically depending on the stream type
gul_header = np.int32(1 | 2 << 24).tobytes()

items_data_type = nb.from_dtype(np.dtype([('item_id', np.int32),
                                          ('damagecdf_i', np.int32),
                                          ('rng_index', np.int32)
                                          ]))

items_MC_data_type = nb.from_dtype(np.dtype([('item_id', np.int32),
                                             ('areaperil_id', areaperil_int),
                                             ('vulnerability_id', np.int32),
                                             ('hazcdf_i', np.int32),
                                             ('rng_index', np.int32),
                                             ('eff_vuln_cdf_i', np.int32),
                                             ('eff_vuln_cdf_Ndamage_bins', np.int32)
                                             ]))

VulnCdfLookup = nb.from_dtype(np.dtype([('start', np.int32), ('length', np.int32)]))


coverage_type = nb.from_dtype(np.dtype([('tiv', np.float64),
                                        ('max_items', np.int32),
                                        ('start_items', np.int32),
                                        ('cur_items', np.int32)
                                        ]))

NP_BASE_ARRAY_SIZE = 8

# negative sidx (definition)
MEAN_IDX = -1
STD_DEV_IDX = -2
TIV_IDX = -3
CHANCE_OF_LOSS_IDX = -4
MAX_LOSS_IDX = -5

NUM_IDX = 5

ITEM_MAP_KEY_TYPE = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int32))
ITEM_MAP_VALUE_TYPE = nb.types.UniTuple(nb.types.int32, 3)
ITEM_MAP_KEY_TYPE_internal = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int64))
ITEM_MAP_VALUE_TYPE_internal = nb.types.UniTuple(nb.types.int32, 2)

# compute the relative size of oasis_float and areaperil_int vs int32
oasis_float_to_int32_size = oasis_float.itemsize // np.int32().itemsize
areaperil_int_to_int32_size = areaperil_int.itemsize // np.int32().itemsize

haz_cdf_type = nb.from_dtype(np.dtype([('probability', oasis_float),
                                       ('intensity_bin_id', np.int32)]))

ProbMean = nb.from_dtype(np.dtype([('prob_to', oasis_float),
                                   ('bin_mean', oasis_float)
                                   ]))
ProbMean_size = ProbMean.size

damagecdfrec_stream = nb.from_dtype(np.dtype([('event_id', np.int32),
                                              ('areaperil_id', areaperil_int),
                                              ('vulnerability_id', np.int32)
                                              ]))

damagecdfrec = nb.from_dtype(np.dtype([('areaperil_id', areaperil_int),
                                       ('vulnerability_id', np.int32)
                                       ]))


gulSampleslevelHeader = nb.from_dtype(np.dtype([('event_id', 'i4'),
                                                ('item_id', 'i4'),
                                                ]))
gulSampleslevelHeader_size = gulSampleslevelHeader.size

gulSampleslevelRec = nb.from_dtype(np.dtype([('sidx', 'i4'),
                                             ('loss', oasis_float),
                                             ]))
gulSampleslevelRec_size = gulSampleslevelRec.size


Keys = {'LocID': np.int32,
        'PerilID': 'category',
        'CoverageTypeID': np.int32,
        'AreaPerilID': areaperil_int,
        'VulnerabilityID': np.int32}



__all__ = [
    'AdminCmd',
    'CreateComplexModelCmd',
    'CreateSimpleModelCmd',
    'EnableBashCompleteCmd',
]


from argparse import RawDescriptionHelpFormatter
from .command import OasisBaseCommand, OasisComputationCommand


class EnableBashCompleteCmd(OasisComputationCommand):
    """
    Adds required command to `.bashrc` Linux or .bash_profile for mac
    so that Command autocomplete works for oasislmf CLI
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'HelperTabComplete'


class CreateSimpleModelCmd(OasisComputationCommand):
    """
    Creates a local Git repository for a "simple model" (using the
    ``cookiecutter`` package) on a "simple model" repository template
    on GitHub)
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'CreateModelRepo'


class CreateComplexModelCmd(OasisComputationCommand):
    """
    Creates a local Git repository for a "complex model" (using the
    ``cookiecutter`` package) on a "simple model" repository template
    on GitHub)
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'CreateComplexModelRepo'


class AdminCmd(OasisBaseCommand):
    """
    Admin subcommands::

        * creates a local Git repository for a "simple model" (using the
          ``cookiecutter`` package) on a "simple model" repository template
          on GitHub)
        * creates a local Git repository for a "complex model" (using the
          ``cookiecutter`` package) on a "complex model" repository template
          on GitHub)
    """
    sub_commands = {
        'create-simple-model': CreateSimpleModelCmd,
        'create-complex-model': CreateComplexModelCmd,
        'enable-bash-complete': EnableBashCompleteCmd
    }



#!/usr/bin/env python

import sys
import argparse
from contextlib import ExitStack

from oasislmf.pytools.gul.manager import read_getmodel_stream

parser = argparse.ArgumentParser(
    usage='use "%(prog)s --help" for more information',
    formatter_class=argparse.RawTextHelpFormatter  # for multi-line help text
)

parser.add_argument('-s', help='skip header (default: False).', default=False, action='store_true', dest='skip_header')
parser.add_argument('--run-dir', help='path to the run directory (default: ".")', default='.')


def print_cdftocsv(event_id, damagecdf, Nbins, rec):
    """Print the cdf produced by getmodel to csv file.
    Note that the input arrays are lists of cdf entries, namely
    the shape on axis=0 is the number of entries.

    Args:
        event_id (int): event_id
        damagecdf (array-like, damagecdf): damage cdf record
        Nbins (array-like, int): number of damage bins
        rec (array-like, rec): cdf record

    Returns:
        list[str]: list of csv lines
    """
    # TODO: accelerate this with numba when it will support string formatting

    # number of cdf entries in the input data
    Nentries = damagecdf.shape[0]

    # build the csv lines
    csv_lines = []
    for i in range(Nentries):
        csv_line_fixed = f"{event_id},{damagecdf[i]['areaperil_id']},{damagecdf[i]['vulnerability_id']},"
        for j in range(Nbins[i]):
            # note that bin index starts from 1 in the csv
            csv_lines.append(csv_line_fixed + f"{j + 1},{rec[i, j]['prob_to']:8.6f},{rec[i, j]['bin_mean']:8.6f}\n")

    return csv_lines


def run(run_dir, skip_header, file_in=None):
    """Run cdftocsv command: convert the cdf output from getmodel into csv format.
    The binary data is read from an input stream, and the csv file is streamed to stdout.

    Args:
        run_dir ([str]): Path to the run directory.

        skip_header ([bool]): If True, does not print the csv header.

    Raises:
        ValueError: If the stream type is not 1.

    """
    with ExitStack() as stack:
        if file_in is None:
            streams_in = sys.stdin.buffer
        else:
            streams_in = stack.enter_context(open(file_in, 'rb'))

        stream_out = sys.stdout

    if not skip_header:
        stream_out.write("event_id,areaperil_id,vulnerability_id,bin_index,prob_to,bin_mean\n")

    # TODO: try using np.savetxt for better performance
    for event_id, damagecdf, Nbins, rec in read_getmodel_stream(run_dir, streams_in):
        lines = print_cdftocsv(event_id, damagecdf, Nbins, rec)
        stream_out.writelines(lines)


def main():
    kwargs = vars(parser.parse_args())

    run(**kwargs)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import argparse
import logging

from oasislmf import __version__ as oasis_version
from oasislmf.pytools.gulmc import logger, manager

parser = argparse.ArgumentParser(
    usage='use "%(prog)s --help" for more information',
    formatter_class=argparse.RawTextHelpFormatter  # for multi-line help text
)

# arguments in alphabetical order (lower-case, then upper-case, then long arguments)
parser.add_argument('-a', help='back-allocation rule. Default: 0', default=0, type=int, dest='alloc_rule')
parser.add_argument('-d', help='output the ground up loss (0), the random numbers used for hazard sampling (1), '
                               'the random numbers used for damage sampling (2). Default: 0',
                    action='store', type=int, dest='debug', default=0)
parser.add_argument('-i', '--file-in', help='filename of input stream (list of events from `eve`).',
                    action='store', type=str, dest='file_in')
parser.add_argument('-o', '--file-out', help='filename of output stream (ground up losses).',
                    action='store', type=str, dest='file_out')
parser.add_argument('-L', help='Loss treshold. Default: 1e-6', default=1e-6,
                    action='store', type=float, dest='loss_threshold')
parser.add_argument('-S', help='Sample size. Default: 0', default=0, action='store', type=int, dest='sample_size')
parser.add_argument('-V', '--version', action='version', version='{}'.format(oasis_version))
parser.add_argument('--effective-damageability',
                    help='if passed true, the effective damageability is used to draw loss samples instead of full MC. Default: False',
                    action='store_true', dest='effective_damageability', default=False)
parser.add_argument('--ignore-correlation',
                    help='if passed true, peril correlation groups (if defined) are ignored for the generation of correlated samples. Default: False',
                    action='store_true', dest='ignore_correlation', default=False)
parser.add_argument('--ignore-file-type', nargs='*',
                    help='the type of file to be loaded. Default: set()', default=set())
parser.add_argument('--data-server', help='=Use tcp/sockets for IPC data sharing.',
                    action='store_true', dest='data_server')
parser.add_argument('--logging-level',
                    help='logging level (debug:10, info:20, warning:30, error:40, critical:50). Default: 30',
                    default=30, type=int)
parser.add_argument('--vuln-cache-size', help='Size in MB of the in-memory cache to store and reuse vulnerability cdf. Default: 200',
                    default=200, action='store', type=int, dest='max_cached_vuln_cdf_size_MB')
parser.add_argument('--peril-filter', help='Id of the peril to keep, if empty take all perils',
                    nargs='+', dest='peril_filter')
parser.add_argument('--random-generator',
                    help='random number generator\n0: numpy default (MT19937), 1: Latin Hypercube. Default: 1',
                    default=1, type=int, dest='random_generator')
parser.add_argument('--run-dir', help='path to the run directory. Default: "."', default='.')


def main():
    # parse arguments to variables
    # note: the long flag name (e.g., '--opt-one') is used as variable name (i.e, the `dest`).
    # hyphens in the long flag name are parsed to underscores, e.g. '--opt-one' is stored in `opt_one``
    kwargs = vars(parser.parse_args())

    # add handler to gul logger
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    manager.run(**kwargs)


if __name__ == '__main__':
    main()



__all__ = [
    'OasisBuiltinBaseLookup',
    'OasisBaseKeysLookup',
    'OasisLookupFactory'
]


# This is a backwards compatibility file for lookup code based on version 1.9.0 and older
# Add deprecated warning, Load equiv classes from ..lookup. *

from ..lookup.factory import KeyServerFactory as OasisLookupFactory
from ..lookup.base import OasisBaseLookup as OasisBuiltinBaseLookup
from ..lookup.interface import OasisLookupInterface as OasisBaseKeysLookup



import argparse
from contextlib import ExitStack
import logging
import mmap
import numpy as np
import os
from zlib import compress

from .getmodel.common import (FootprintHeader, EventIndexBin, EventIndexBinZ,
                              footprint_filename, footprint_index_filename, zfootprint_filename, zfootprint_index_filename)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def compress_footprint(static_path, decompressed_size=True, compression_level=-1):
    with ExitStack() as stack:
        footprint_obj = stack.enter_context(open(os.path.join(static_path, footprint_filename), 'rb'))
        footprint_map = mmap.mmap(footprint_obj.fileno(), length=0, access=mmap.ACCESS_READ)
        footprint_header = np.frombuffer(bytearray(footprint_map[:FootprintHeader.size]), dtype=FootprintHeader)
        if decompressed_size:
            footprint_header['has_intensity_uncertainty'] |= 2  # set compressed byte of has_intensity_uncertainty to 1
            index_dtype = EventIndexBinZ
        else:
            index_dtype = EventIndexBin

        zfootprint_obj = stack.enter_context(open(os.path.join(static_path, zfootprint_filename), 'wb'))
        zfootprint_obj.write(footprint_header.tobytes())

        footprint_idx_bin = np.memmap(os.path.join(static_path, footprint_index_filename), dtype=EventIndexBin, mode='r')
        zfootprint_idx_bin = np.memmap(os.path.join(static_path, zfootprint_index_filename), dtype=index_dtype, mode='w+',
                                       shape=footprint_idx_bin.shape)

        offset = 8
        in_size = 8
        for i in range(footprint_idx_bin.shape[0]):
            index = footprint_idx_bin[i]
            zindex = zfootprint_idx_bin[i]
            zindex['event_id'] = index['event_id']
            zindex['offset'] = offset
            if decompressed_size:
                zindex['d_size'] = index['size']

            data_in = footprint_map[index['offset']: index['offset'] + index['size']]
            data_out = compress(data_in, compression_level)
            zindex['size'] = len(data_out)
            offset += len(data_out)
            in_size += index['size']
            zfootprint_obj.write(data_out)

        logger.info(f'compressed {os.path.join(static_path, footprint_filename)} from {in_size} B => {offset} B')


parser = argparse.ArgumentParser()
parser.add_argument('-p', '--static-path', help='path to the footprint', default='.')
parser.add_argument('-l', '--compression-level', help='integer from 0 to 9 or -1 controlling the level of compression',
                    default=-1, type=int)
parser.add_argument('--decompressed_size', help='if True add the decompressed size to the index file', action='store_true')
parser.add_argument('-v', '--logging-level', help='logging level (debug:10, info:20, warning:30, error:40, critical:50)',
                    default=20, type=int)


def footprintconvpy():
    kwargs = vars(parser.parse_args())

    # add handler to fm logger
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    compress_footprint(**kwargs)


if __name__ == '__main__':
    footprintconvpy()



"""
TODO: should I check values are valid in the financial structure percentage not between 0 and 1 (ex: deductible, limit ...)
TODO: validate max and min ded implementation
TODO: It seems that if a policy with share is used, subsequent policy using min or max deductible will be wrong
     so it make no sense to compute deductible, over_limit, under_limit

"""


from numba import njit


class UnknownCalcrule(Exception):
    pass


@njit(cache=True)
def min2(a, b):
    return a if a < b else b


@njit(cache=True, fastmath=True)
def calcrule_1(policy, loss_out, loss_in):
    """
    Deductible and limit
    """
    lim = policy['limit_1'] + policy['deductible_1']
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['deductible_1']:
            loss_out[i] = 0
        elif loss_in[i] <= lim:
            loss_out[i] = loss_in[i] - policy['deductible_1']
        else:
            loss_out[i] = policy['limit_1']


@njit(cache=True, fastmath=True)
def calcrule_2(policy, loss_out, loss_in):
    """
    Deductible, attachment, limit and share

    """
    ded_att = policy['deductible_1'] + policy['attachment_1']
    lim = policy['limit_1'] + ded_att
    maxi = policy['limit_1'] * policy['share_1']
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= ded_att:
            loss_out[i] = 0
        elif loss_in[i] <= lim:
            loss_out[i] = (loss_in[i] - ded_att) * policy['share_1']
        else:
            loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_3(policy, loss_out, loss_in):
    """
    Franchise deductible and limit
    """
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['deductible_1']:
            loss_out[i] = 0
        elif loss_in[i] <= policy['limit_1']:
            loss_out[i] = loss_in[i]
        else:
            loss_out[i] = policy['limit_1']


@njit(cache=True, fastmath=True)
def calcrule_5(policy, loss_out, loss_in):
    """
    Deductible and limit as a proportion of loss
    """
    effective_deductible = loss_in * policy['deductible_1']
    effective_limit = loss_in * policy['limit_1']
    if policy['deductible_1'] + policy['limit_1'] >= 1:  # always under limit
        for i in range(loss_in.shape[0]):
            loss_out[i] = loss_in[i] - effective_deductible[i]

    else:  # always over limit
        loss_out[:] = effective_limit


@njit(cache=True, fastmath=True)
def calcrule_12(policy, loss_out, loss_in):
    """
    Deductible only
    """
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['deductible_1']:
            loss_out[i] = 0
        else:
            loss_out[i] = loss_in[i] - policy['deductible_1']


@njit(cache=True, fastmath=True)
def calcrule_14(policy, loss_out, loss_in):
    """
    Limit only
    """
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['limit_1']:
            loss_out[i] = loss_in[i]
        else:
            loss_out[i] = policy['limit_1']


@njit(cache=True, fastmath=True)
def calcrule_15(policy, loss_out, loss_in):
    """
    deductible and limit % loss
    """
    effective_limit = policy['deductible_1'] / (1 - policy['limit_1'])
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['deductible_1']:
            loss_out[i] = 0
        elif loss_in[i] <= effective_limit:
            loss_out[i] = loss_in[i] - policy['deductible_1']
        else:
            loss_out[i] = loss_in[i] * policy['limit_1']


@njit(cache=True, fastmath=True)
def calcrule_16(policy, loss_out, loss_in):
    """
    deductible % loss
    """
    loss_out[:] = loss_in * (1 - policy['deductible_1'])


@njit(cache=True, fastmath=True)
def calcrule_17(policy, loss_out, loss_in):
    """
    deductible % loss with attachment, limit and share
    """
    if policy['deductible_1'] >= 1:
        loss_out.fill(0)
    else:
        post_ded_attachment = policy['attachment_1'] / (1 - policy['deductible_1'])
        post_ded_attachment_limit = (policy['attachment_1'] + policy['limit_1']) / (1 - policy['deductible_1'])
        maxi = policy['limit_1'] * policy['share_1']
        for i in range(loss_in.shape[0]):
            if loss_in[i] <= post_ded_attachment:
                loss_out[i] = 0
            elif loss_in[i] <= post_ded_attachment_limit:
                loss_out[i] = (loss_in[i] * (1 - policy['deductible_1']) - policy['attachment_1']) * policy['share_1']
            else:
                loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_20(policy, loss_out, loss_in):
    """
    reverse franchise deductible
    """
    for i in range(loss_in.shape[0]):
        if loss_in[i] > policy['deductible_1']:
            loss_out[i] = 0
        else:
            loss_out[i] = loss_in[i]


@njit(cache=True, fastmath=True)
def calcrule_22(policy, loss_out, loss_in):
    """
    reinsurance % ceded, limit and % placed
    """
    if policy['share_1'] == 0:
        loss_out.fill(0)
    else:
        pre_share_limit = policy['limit_1'] / policy['share_1']
        all_share = policy['share_1'] * policy['share_2'] * policy['share_3']
        maxi = policy['limit_1'] * policy['share_2'] * policy['share_3']
        for i in range(loss_in.shape[0]):
            if loss_in[i] <= pre_share_limit:
                loss_out[i] = loss_in[i] * all_share
            else:
                loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_23(policy, loss_out, loss_in):
    """
    reinsurance limit and % placed
    """
    all_share = policy['share_2'] * policy['share_3']
    maxi = policy['limit_1'] * all_share
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['limit_1']:
            loss_out[i] = loss_in[i] * all_share
        else:
            loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_24(policy, loss_out, loss_in):
    """
    reinsurance excess terms
    """
    if policy['share_1'] == 0:
        loss_out.fill(0)
    else:
        pre_share_attachment = policy['attachment_1'] / policy['share_1']
        pre_share_attachment_limit = (policy['limit_1'] + policy['attachment_1']) / policy['share_1']
        attachment_share = policy['attachment_1'] * policy['share_2'] * policy['share_3']
        all_share = policy['share_1'] * policy['share_2'] * policy['share_3']
        maxi = policy['limit_1'] * policy['share_2'] * policy['share_3']
        for i in range(loss_in.shape[0]):
            if loss_in[i] <= pre_share_attachment:
                loss_out[i] = 0
            elif loss_in[i] <= pre_share_attachment_limit:
                loss_out[i] = loss_in[i] * all_share - attachment_share
            else:
                loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_25(policy, loss_out, loss_in):
    """
    reinsurance proportional terms
    """
    loss_out[:] = loss_in * (policy['share_1'] * policy['share_2'] * policy['share_3'])


@njit(cache=True, fastmath=True)
def calcrule_28(policy, loss_out, loss_in):
    """
    % loss step payout
    note that 1 was added to scale_1 in the precompute step
    """
    if policy['step_id'] == 1:
        loss_out.fill(0)
    for i in range(loss_in.shape[0]):
        if policy['trigger_start'] <= loss_in[i] < policy['trigger_end']:
            loss = max(policy['payout_start'] * loss_in[i] - policy['deductible_1'], 0)
            loss_out[i] = (loss + min(loss * policy['scale_2'], policy['limit_2'])) * policy['scale_1']


@njit(cache=True, fastmath=True)
def calcrule_281(policy, loss_out, loss_in):
    """
    conditional coverage
    """
    if policy['step_id'] == 1:
        loss_out.fill(0)
    for i in range(loss_in.shape[0]):
        if policy['trigger_start'] <= loss_in[i] < policy['trigger_end']:
            loss_out[i] += min(loss_out[i] * policy['scale_2'], policy['limit_2']) * policy['scale_1']


@njit(cache=True, fastmath=True)
def calcrule_32(policy, loss_out, loss_in):
    """
    monetary amount trigger and % loss step payout with limit
    """
    if policy['step_id'] == 1:
        loss_out.fill(0)
    for i in range(loss_in.shape[0]):
        if policy['trigger_start'] <= loss_in[i]:
            loss = min(policy['payout_start'] * loss_in[i], policy['limit_1'])
            loss_out[i] += (loss + min(loss * policy['scale_2'], policy['limit_2'])) * policy['scale_1']


@njit(cache=True, fastmath=True)
def calcrule_33(policy, loss_out, loss_in):
    """
    deductible % loss with limit

    """
    if policy['deductible_1'] >= 1:
        loss_out.fill(0)
    else:
        post_ded_limit = policy['limit_1'] / (1 - policy['deductible_1'])
        for i in range(loss_in.shape[0]):
            if loss_in[i] <= post_ded_limit:
                loss_out[i] = loss_in[i] * (1 - policy['deductible_1'])
            else:
                loss_out[i] = policy['limit_1']


@njit(cache=True, fastmath=True)
def calcrule_34(policy, loss_out, loss_in):
    """
    deductible with attachment and share

    TODO: compare to the cpp, as there is shares, deductible won't be use later on so no need to compute it
    """
    ded_att = policy['deductible_1'] + policy['attachment_1']
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= ded_att:
            loss_out[i] = 0
        else:
            loss_out[i] = (loss_in[i] - ded_att) * policy['share_1']


@njit(cache=True, fastmath=True)
def calcrule_37(policy, loss_out, loss_in):
    """
    % loss step payout
    """
    if policy['step_id'] == 1:
        loss_out.fill(0)
    for i in range(loss_in.shape[0]):
        if policy['trigger_start'] <= loss_in[i] < policy['trigger_end']:
            loss = min(max(policy['payout_start'] * loss_in[i] - policy['deductible_1'], 0), policy['limit_1'])
            loss_out[i] = (loss + min(loss * policy['scale_2'], policy['limit_2'])) * policy['scale_1']


@njit(cache=True, fastmath=True)
def calcrule_38(policy, loss_out, loss_in):
    """
    conditional coverage
    """
    if policy['step_id'] == 1:
        loss_out.fill(0)
    for i in range(loss_in.shape[0]):
        if policy['trigger_start'] <= loss_in[i] < policy['trigger_end']:
            loss_out[i] = (loss_out[i] + min(loss_out[i] * policy['scale_2'], policy['limit_2'])) * (policy['scale_1'])


@njit(cache=True)
def calc(policy, loss_out, loss_in, stepped):
    if policy['calcrule_id'] == 1:
        calcrule_1(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 2:
        calcrule_2(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 3:
        calcrule_3(policy, loss_out, loss_in)
    # calcrule_4 (deductible % TIV and limit) is redirected to 1 when building financial structure
    elif policy['calcrule_id'] == 5:
        calcrule_5(policy, loss_out, loss_in)
    # calcrule_6 (deductible % TIV) is redirected to 12 when building financial structure
    # calcrule_9 (limit with deductible % limit) is redirected to 1 when building financial structure
    elif policy['calcrule_id'] == 12:
        calcrule_12(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 14:
        calcrule_14(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 15:
        calcrule_15(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 16:
        calcrule_16(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 17:
        calcrule_17(policy, loss_out, loss_in)
    # calcrule_18 (deductible % tiv with attachment, limit and share) is redirected to 2 when building financial structure
    elif policy['calcrule_id'] == 20:
        calcrule_20(policy, loss_out, loss_in)
    # calcrule_21 (deductible % tiv with min and max deductible) is redirected to 13 when building financial structure
    elif policy['calcrule_id'] == 22:
        calcrule_22(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 23:
        calcrule_23(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 24:
        calcrule_24(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 25:
        calcrule_25(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 33:
        calcrule_33(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 34:
        calcrule_34(policy, loss_out, loss_in)
    elif policy['calcrule_id'] == 100:
        loss_out[:] = loss_in
    elif stepped is not None:
        if policy['calcrule_id'] == 28:
            calcrule_28(policy, loss_out, loss_in)
        elif policy['calcrule_id'] == 32:
            calcrule_32(policy, loss_out, loss_in)
        elif policy['calcrule_id'] == 37:
            calcrule_37(policy, loss_out, loss_in)
        elif policy['calcrule_id'] == 38:
            calcrule_38(policy, loss_out, loss_in)
        else:
            raise UnknownCalcrule()
    else:
        raise UnknownCalcrule()



__all__ = [
    'ReinsuranceLayer',
    'write_files_for_reinsurance'
]

import logging
import os
import shutil

from collections import namedtuple

import numbers
import pandas as pd

from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from . import oed

from ..utils.data import get_dataframe

# Metadata about an inuring layer
InuringLayer = namedtuple(
    "InuringLayer",
    "inuring_priority reins_numbers is_valid validation_messages"
)

RiInputs = namedtuple(
    'RiInputs',
    'inuring_priority risk_level ri_inputs'
)

RiLayerInputs = namedtuple(
    'RiLayerInputs',
    'fm_programme fm_profile fm_policytc'
)


def _get_location_tiv(location, coverage_type_id):
    0 / 0
    switcher = {
        oed.BUILDING_COVERAGE_TYPE_ID: location.get('BuildingTIV', 0),
        oed.OTHER_BUILDING_COVERAGE_TYPE_ID: location.get('OtherTIV', 0),
        oed.CONTENTS_COVERAGE_TYPE_ID: location.get('ContentsTIV', 0),
        oed.TIME_COVERAGE_TYPE_ID: location.get('BITIV', 0)
    }
    return switcher.get(coverage_type_id, 0)


def _get_ri_inputs(
        items_df,
        coverages_df,
        xref_descriptions_df,
        ri_info_df,
        ri_scope_df):
    ri_inputs = []
    for inuring_priority in range(1, ri_info_df['InuringPriority'].max() + 1):
        # Filter the reinsNumbers by inuring_priority
        reins_numbers = ri_info_df[ri_info_df['InuringPriority'] == inuring_priority].ReinsNumber.tolist()
        risk_level_set = set(ri_info_df[ri_info_df['ReinsNumber'].isin(reins_numbers)].RiskLevel)

        for risk_level in oed.REINS_RISK_LEVELS:
            if risk_level not in risk_level_set:
                continue

            ri_inputs.append(
                RiInputs(
                    inuring_priority=inuring_priority,
                    risk_level=risk_level,
                    ri_inputs=_generate_inputs_for_reinsurance_risk_level(
                        inuring_priority,
                        items_df,
                        coverages_df,
                        xref_descriptions_df,
                        ri_info_df,
                        ri_scope_df,
                        risk_level)))
    return ri_inputs


@oasis_log
def write_files_for_reinsurance(
        gul_inputs_df,
        xref_descriptions_df,
        ri_info_df,
        ri_scope_df,
        fm_xref_fp,
        output_dir,
        store_tree=False):
    """
    Generate files for reinsurance.
    """
    inuring_metadata = {}

    items_df = gul_inputs_df.loc[:, ['item_id', 'coverage_id', 'areaperil_id', 'vulnerability_id', 'group_id']].drop_duplicates()
    coverages_df = gul_inputs_df.loc[:, ['coverage_id', 'tiv']].drop_duplicates()
    ri_inputs = _get_ri_inputs(
        items_df,
        coverages_df,
        xref_descriptions_df,
        ri_info_df,
        ri_scope_df)

    reinsurance_index = 1

    for ri_input in ri_inputs:

        ri_output_dir = os.path.join(output_dir, "RI_{}".format(reinsurance_index))
        if os.path.exists(ri_output_dir):
            shutil.rmtree(ri_output_dir)
        os.makedirs(ri_output_dir)

        ri_input.ri_inputs.fm_programme.to_csv(
            os.path.join(ri_output_dir, "fm_programme.csv"), index=False)
        ri_input.ri_inputs.fm_profile.to_csv(
            os.path.join(ri_output_dir, "fm_profile.csv"), index=False)
        ri_input.ri_inputs.fm_policytc.to_csv(
            os.path.join(ri_output_dir, "fm_policytc.csv"), index=False)

        fm_xref_df = get_dataframe(fm_xref_fp)
        fm_xref_df['agg_id'] = range(1, 1 + len(fm_xref_df))
        # Net losses across all layers is associated to the max layer ID.
        fm_xref_df['layer_id'] = ri_input.ri_inputs.fm_policytc['layer_id'].max()
        fm_xref_df.to_csv(
            os.path.join(ri_output_dir, "fm_xref.csv"), index=False)

        inuring_metadata[reinsurance_index] = {
            'inuring_priority': ri_input.inuring_priority,
            'risk_level': ri_input.risk_level,
            'directory': ri_output_dir
        }

        reinsurance_index = reinsurance_index + 1

    return inuring_metadata


def _generate_inputs_for_reinsurance_risk_level(
        inuring_priority,
        items_df,
        coverages_df,
        xref_descriptions_df,
        ri_info_df,
        ri_scope_df,
        risk_level):
    """
    Generate files for a reinsurance risk level.
    """
    reins_numbers_1 = ri_info_df[
        ri_info_df['InuringPriority'] == inuring_priority].ReinsNumber
    if reins_numbers_1.empty:
        return None
    reins_numbers_2 = ri_info_df[
        ri_info_df.isin({"ReinsNumber": reins_numbers_1.tolist()}).ReinsNumber
        & (ri_info_df.RiskLevel == risk_level)
    ].ReinsNumber
    if reins_numbers_2.empty:
        return None

    ri_info_inuring_priority_df = ri_info_df[ri_info_df.isin(
        {"ReinsNumber": reins_numbers_2.tolist()}).ReinsNumber]
    output_name = "ri_{}_{}".format(inuring_priority, risk_level)
    reinsurance_layer = ReinsuranceLayer(
        name=output_name,
        ri_info_df=ri_info_inuring_priority_df,
        ri_scope_df=ri_scope_df,
        items_df=items_df,
        coverages_df=coverages_df,
        xref_descriptions_df=xref_descriptions_df,
        risk_level=risk_level
    )

    reinsurance_layer.generate_oasis_structures()
    return RiLayerInputs(
        fm_programme=reinsurance_layer.fmprogrammes_df,
        fm_profile=reinsurance_layer.fmprofiles_df,
        fm_policytc=reinsurance_layer.fm_policytcs_df
    )


class ReinsuranceLayer(object):
    """
    Generates ktools inputs and runs financial module for a reinsurance structure.
    """

    def __init__(
            self,
            name, ri_info_df, ri_scope_df, items_df, coverages_df,
            xref_descriptions_df, risk_level, fmsummaryxref_df=pd.DataFrame(),
            gulsummaryxref_df=pd.DataFrame(), logger=None
    ):

        self.logger = logger or logging.getLogger(__name__)
        self.name = name

        self.coverages_df = coverages_df
        self.items_df = items_df
        self.xref_descriptions_df = xref_descriptions_df
        self.fmsummaryxref_df = fmsummaryxref_df
        self.gulsummaryxref_df = gulsummaryxref_df

        self.item_ids = list()
        self.item_tivs = list()
        self.fmprogrammes_df = pd.DataFrame()
        self.fmprofiles_df = pd.DataFrame()
        self.fm_policytcs_df = pd.DataFrame()

        self.risk_level = risk_level

        self.ri_info_df = ri_info_df
        self.ri_scope_df = ri_scope_df

    def _is_valid_id(self, id_to_check):
        is_valid = (
            self._is_defined(id_to_check) and
            (
                (isinstance(id_to_check, str) and id_to_check != "") or
                (isinstance(id_to_check, numbers.Number) and id_to_check > 0)
            )
        )
        return is_valid

    def _get_reins_type_fields(self, risk_level=None, all_fields=False):

        if all_fields:
            return ['PortNumber', 'AccNumber', 'PolNumber', 'LocGroup', 'LocNumber']

        if risk_level is None:
            risk_level = self.risk_level

        fields = []
        if risk_level == oed.REINS_RISK_LEVEL_PORTFOLIO:
            fields = ['PortNumber']
        elif risk_level == oed.REINS_RISK_LEVEL_ACCOUNT:
            fields = ['PortNumber', 'AccNumber']
        elif risk_level == oed.REINS_RISK_LEVEL_POLICY:
            fields = ['PortNumber', 'AccNumber', 'PolNumber']
        elif risk_level == oed.REINS_RISK_LEVEL_LOCATION_GROUP:
            fields = ['LocGroup']
        elif risk_level == oed.REINS_RISK_LEVEL_LOCATION:
            fields = ['PortNumber', 'AccNumber', 'LocNumber']
        else:
            raise OasisException(f"Unknown risk level: {risk_level}")
        return fields

    def _get_valid_fields(self, fields):

        valid_fields = []
        for field in fields:
            valid_field = field + '_valid'
            valid_fields.append(valid_field)
        return valid_fields

    def _match_exact(self, df_, ri_df_, fields=None):

        if fields == None:
            fields = []

        df_ = df_.merge(
            ri_df_[fields + ['layer_id', 'level_id', 'profile_id']].drop_duplicates(),
            how='left', on=fields + ['layer_id', 'level_id'], suffixes=['', '_y']
        )
        df_['profile_id'] = df_['profile_id'].where(
            df_['profile_id_y'].isna(), df_['profile_id_y']
        )

        return df_['profile_id'].to_list()

    def _match_portfolio(self, df_, ri_df_):

        fields = self._get_reins_type_fields(oed.REINS_RISK_LEVEL_PORTFOLIO)
        valid_fields = self._get_valid_fields(fields)

        df_['profile_id'] = self._match_exact(
            df_=df_, ri_df_=ri_df_[ri_df_[valid_fields[0]] == True],
            fields=fields
        )

        df_['profile_id'] = self._match_exact(
            df_=df_, ri_df_=ri_df_[ri_df_[valid_fields[0]] == False]
        )

        return df_['profile_id'].to_list()

    def _match_account(self, df_, ri_df_):

        fields = self._get_reins_type_fields(oed.REINS_RISK_LEVEL_ACCOUNT)
        valid_fields = self._get_valid_fields(fields)

        df_['profile_id'] = self._match_exact(
            df_=df_,
            ri_df_=ri_df_[(ri_df_[valid_fields[0]] == True) & (ri_df_[valid_fields[1]] == True)],
            fields=fields
        )

        df_['profile_id'] = self._match_portfolio(
            df_=df_, ri_df_=ri_df_[ri_df_[valid_fields[1]] == False]
        )

        return df_['profile_id'].to_list()

    def _match_policy_or_location(self, df_, ri_df_):

        fields = self._get_reins_type_fields()
        valid_fields = self._get_valid_fields(fields)

        df_['profile_id'] = self._match_exact(
            df_=df_,
            ri_df_=ri_df_[(ri_df_[valid_fields[0]] == True) & (ri_df_[valid_fields[1]] == True) & (ri_df_[valid_fields[2]] == True)],
            fields=fields
        )

        df_['profile_id'] = self._match_account(
            df_=df_, ri_df_=ri_df_[ri_df_[valid_fields[2]] == False]
        )

        return df_['profile_id'].to_list()

    def _match_location_group(self, df_, ri_df_):

        fields = self._get_reins_type_fields()
        valid_fields = self._get_valid_fields(fields)

        df_['profile_id'] = self._match_exact(
            df_=df_, ri_df_=ri_df_[ri_df_[valid_fields[0]] == True],
            fields=fields
        )

        return df_['profile_id'].to_list()

    def _match_row(self, row):
        match = True
        if match and row['PortNumber_valid']:
            match = row['PortNumber_x'] == row['PortNumber_y']
        if match and row['AccNumber_valid']:
            match = row['AccNumber_x'] == row['AccNumber_y']
        if match and row['PolNumber_valid']:
            match = row['PolNumber_x'] == row['PolNumber_y']
        if match and row['LocGroup_valid']:
            match = row['LocGroup_x'] == row['LocGroup_y']
        if match and row['LocNumber_valid']:
            match = row['LocNumber_x'] == row['LocNumber_y']

        return match

    def _get_risk_level_profile_ids(self, df_, ri_df_, exact=False):
        profile_ids = []
        if df_.empty:
            return profile_ids

        if exact:
            profile_ids = self._match_exact(
                df_=df_, ri_df_=ri_df_,
                fields=self._get_reins_type_fields()
            )
        elif self.risk_level == oed.REINS_RISK_LEVEL_PORTFOLIO:
            profile_ids = self._match_portfolio(df_, ri_df_)
        elif self.risk_level == oed.REINS_RISK_LEVEL_ACCOUNT:
            profile_ids = self._match_account(df_, ri_df_)
        elif self.risk_level == oed.REINS_RISK_LEVEL_POLICY:
            profile_ids = self._match_policy_or_location(df_, ri_df_)
        elif self.risk_level == oed.REINS_RISK_LEVEL_LOCATION:
            profile_ids = self._match_policy_or_location(df_, ri_df_)
        elif self.risk_level == oed.REINS_RISK_LEVEL_LOCATION_GROUP:
            profile_ids = self._match_location_group(df_, ri_df_)
        else:
            raise OasisException(f"Unknown risk level: {self.risk_level}")

        return profile_ids

    def _get_filter_level_profile_ids(self, df_, ri_df_):

        if len(df_) == 0 or len(ri_df_) == 0:
            return []

        df_['idx'] = df_.index
        df_ = df_.merge(ri_df_, how='inner', on='layer_id')
        df_['match'] = df_.apply(lambda row: self._match_row(row), axis=1)
        filter_idx = df_[df_['match'] == True]['idx'].drop_duplicates().to_list()

        return filter_idx

    def _is_defined(self, num_to_check):
        # If the value = NaN it will return False
        return num_to_check == num_to_check

    def _check_ri_df_row(self, row):
        # For some treaty types the scope filter must match exactly
        okay = True
        if row.ReinsType == oed.REINS_TYPE_FAC or row.ReinsType == oed.REINS_TYPE_SURPLUS_SHARE:
            if row.RiskLevel == oed.REINS_RISK_LEVEL_ACCOUNT:
                okay = self._is_valid_id(row.AccNumber) and not self._is_valid_id(row.PolNumber) and not self._is_valid_id(row.LocNumber)
            elif row.RiskLevel == oed.REINS_RISK_LEVEL_POLICY:
                okay = self._is_valid_id(row.AccNumber) and self._is_valid_id(row.PolNumber) and not self._is_valid_id(row.LocNumber)
            elif row.RiskLevel == oed.REINS_RISK_LEVEL_LOCATION:
                okay = self._is_valid_id(row.AccNumber) and self._is_valid_id(row.LocNumber)
            elif row.RiskLevel == oed.REINS_RISK_LEVEL_LOCATION_GROUP:
                okay = self._is_valid_id(row.LocGroup)
        return okay

    def _log_dataframe(self, df_dict):
        container_name = 'df_dict'
        if isinstance(df_dict, dict):
            container_name += '.items()'

        for df_name, df_ in eval(container_name):
            self.logger.debug(f'{df_name}: {self.name}:')
            self.logger.debug(df_)

    LOCATION_RISK_LEVEL = 2

    def _get_xref_df(self):
        """
        Build the cross-reference dataframe, which serves as a representation
        of the insurance programme depending on the reinsurance risk level.
        Dataframes for programme, risk, filter and items levels are created.
        The fields agg_id, level_id and to_agg_id (agg_id_to), which are used
        to construct the FM Programmes structure, are assigned. The
        aforementioned dataframes are concatenated to form a single dataframe
        called xref_df, which is returned. The returned dataframe features the
        fields necessary for the assignment of profile IDs.
        """
        risk_level_id = self.LOCATION_RISK_LEVEL + 1
        program_node_level_id = risk_level_id + 1

        if self.risk_level == oed.REINS_RISK_LEVEL_LOCATION_GROUP:
            xref_descriptions = self.xref_descriptions_df.sort_values(
                by=["LocGroup", "PortNumber", "AccNumber", "PolNumber", "LocNumber"])
        elif self.risk_level == oed.REINS_RISK_LEVEL_LOCATION:
            xref_descriptions = self.xref_descriptions_df.sort_values(
                by=["PortNumber", "AccNumber", "LocNumber", "PolNumber"])
        else:
            xref_descriptions = self.xref_descriptions_df.sort_values(
                by=["PortNumber", "AccNumber", "PolNumber", "LocNumber"])

        df_levels = []
        # Programme level
        programme_level_df = pd.DataFrame(
            {'agg_id': 1, 'level_id': program_node_level_id, 'agg_id_to': 0},
            index=[0]
        )
        df_levels.append('programme_level')

        # Risk level
        fields = self._get_reins_type_fields()
        risk_level_df = pd.DataFrame()
        risk_level_df = xref_descriptions.drop_duplicates(
            subset=fields, keep='first'
        ).reset_index(drop=True)
        risk_level_df['agg_id'] = risk_level_df.index + 1
        risk_level_df['level_id'] = risk_level_id
        risk_level_df['agg_id_to'] = 1
        df_levels.append('risk_level')

        # Filter level
        filter_level_fields = [
            'PortNumber', 'AccNumber', 'PolNumber', 'LocNumber', 'LocGroup'
        ]
        filter_level_df = xref_descriptions.drop_duplicates(
            subset=filter_level_fields
        ).reset_index(drop=True)
        filter_level_df['agg_id'] = filter_level_df.index + 1
        filter_level_df['level_id'] = 2
        filter_level_df = filter_level_df.merge(
            risk_level_df[fields + ['agg_id']], how='left', on=fields,
            suffixes=['', '_to']
        )
        df_levels.append('filter_level')

        # Item level
        item_level_df = xref_descriptions.reset_index(drop=True)
        item_level_df['agg_id'] = item_level_df['output_id']
        item_level_df['level_id'] = 1
        item_level_df = item_level_df.merge(
            filter_level_df[filter_level_fields + ['agg_id']], how='left',
            on=filter_level_fields, suffixes=['', '_to']
        )
        df_levels.append('items_level')

        df_list = [
            programme_level_df, risk_level_df, filter_level_df, item_level_df
        ]

        if self.logger:
            self._log_dataframe(zip(df_levels, df_list))

        xref_df = pd.concat(df_list, ignore_index=True)

        return xref_df

    def _get_risk_level_id(self):
        risk_level_id = 3
        return risk_level_id

    def _get_filter_level_id(self):
        risk_level_id = 2
        return risk_level_id

    def _get_new_profiles_dataframe(self, profiles, fmprofiles_df):
        profiles_df = pd.DataFrame(
            [
                [
                    profile_id, calcrule_id, deductible1, deductible2,
                    deductible3, attachment, limit, share1, share2, share3
                ] for profile_id, calcrule_id, deductible1, deductible2,
                deductible3, attachment, limit, share1, share2,
                share3 in profiles.values
            ],
            columns=profiles.values[0]._fields
        )
        fmprofiles_df = pd.concat(
            [fmprofiles_df, profiles_df], ignore_index=True
        )

        return fmprofiles_df

    def _get_profiles(
            self, ri_df_, fmprofiles_df, attachment='None', limit='None',
            ceded='None', placement='None'
    ):
        profiles = ri_df_.apply(
            lambda row: oed.get_reinsurance_profile(
                row['profile_id'], attachment=eval(attachment),
                limit=eval(limit), ceded=eval(ceded), placement=eval(placement)
            ), axis=1
        ).drop_duplicates()
        if len(profiles) > 0:
            fmprofiles_df = self._get_new_profiles_dataframe(
                profiles, fmprofiles_df
            )

        return fmprofiles_df

    def _get_occlimit_profiles(
            self, occlimit_df_, fmprofiles_df, limit='None', placement='None'
    ):
        profiles = occlimit_df_.apply(
            lambda row: oed.get_occlim_profile(
                row['profile_id'], limit=eval(limit), placement=eval(placement)
            ), axis=1
        ).drop_duplicates()
        if len(profiles) > 0:
            fmprofiles_df = self._get_new_profiles_dataframe(
                profiles, fmprofiles_df
            )

        return fmprofiles_df

    def generate_oasis_structures(self):
        '''
        Create the Oasis structures - FM Programmes, FM Profiles and FM Policy
        TCs - that represent the reinsurance structure.

        The cross-reference dataframe, which serves as a representation of the
        insurance programme depending on the reinsurance risk level, is built.
        With the exception of facultative contracts, each contract is a
        separate layer. Profile IDs for the risk and filter levels are created
        using the merged reinsurance scope and info dataframes. These profile
        IDs are assigned according to some combination of the fields
        PortNumber, AccNumber, PolNumber, LocGroup and LocNumber, dependent on
        reinsurance risk level. Individual programme level profile IDs are
        assigned for each row of the reinsurance info dataframe. Finally, the
        Oasis structure is written out.
        '''

        fmprofiles_list = list()

        profile_id = 1
        nolossprofile_id = profile_id
        fmprofiles_list.append(
            oed.get_no_loss_profile(nolossprofile_id))
        profile_id = profile_id + 1
        passthroughprofile_id = profile_id
        fmprofiles_list.append(
            oed.get_pass_through_profile(passthroughprofile_id))
        fmprofiles_df = pd.DataFrame(fmprofiles_list)

        node_layer_profile_map = {}

        self.logger.debug('No loss and passthrough profiles:')
        self.logger.debug(fmprofiles_list)

        # Get cross-reference dataframe which shall be used to build profile map
        # FM Programmes fields agg_id, level_id and to_agg_id are assigned here
        xref_df = self._get_xref_df()

        # Assign default profile IDs
        xref_df['profile_id'] = nolossprofile_id
        xref_df['profile_id'] = xref_df['profile_id'].where(
            (xref_df['level_id'] == self._get_filter_level_id()) | (xref_df['level_id'] == self._get_risk_level_id()),
            passthroughprofile_id
        )

        # Merge RI info and scope dataframes, and assign layers
        # Use as few layers as possible for FAC
        # Otherwise separate layers for each contract
        self.logger.debug(
            'Merging RI info and scope dataframes and assigning layers'
        )
        fields = self._get_reins_type_fields()
        ri_df = self.ri_info_df[self.ri_info_df['RiskLevel'] == self.risk_level].merge(
            self.ri_scope_df,
            on='ReinsNumber', suffixes=['', '_scope']
        )
        ri_df['layer_id'] = 0
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_FAC, 'layer_id'] = (ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_FAC]
                                                                           .groupby(fields, observed=True).cumcount() + 1)
        ri_info_no_fac = self.ri_info_df[self.ri_info_df['ReinsType'] != oed.REINS_TYPE_FAC].reset_index(drop=True)
        ri_info_no_fac['layer_id'] = ri_info_no_fac.index + 1 + ri_df['layer_id'].max()
        ri_df = ri_df.merge(ri_info_no_fac, how='left', on=ri_info_no_fac.columns.to_list()[:-1], suffixes=['', '_y'])
        ri_df['layer_id'] = ri_df['layer_id'].where(ri_df['layer_id_y'].isna(), ri_df['layer_id_y'])
        ri_df = ri_df.drop('layer_id_y', axis=1)
        del ri_info_no_fac
        ri_df['valid_row'] = ri_df.apply(lambda row: self._check_ri_df_row(row), axis=1)
        if not ri_df['valid_row'].all():
            raise OasisException(
                f'Invalid combination of Risk Level and Reinsurance Type. Please check scope file:\n{ri_df[ri_df["valid_row"] == False]}'
            )
        ri_df = ri_df.drop('valid_row', axis=1)

        profile_maps = [xref_df.copy() for i in range(ri_df['layer_id'].max())]
        for i, df in enumerate(profile_maps):
            df['layer_id'] = i + 1
        profile_map_df = pd.concat(profile_maps, ignore_index=True)

        # Create risk level and filter level profile IDs
        ri_df['level_id'] = 0
        ri_df['profile_id'] = 0
        # Facultative profile IDs
        self.logger.debug('Creating risk level and filter level profile IDs:')
        self.logger.debug(f'{oed.REINS_TYPE_FAC} profiles...')
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_FAC, 'level_id'] = self._get_risk_level_id()
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_FAC, 'profile_id'] = pd.factorize(pd._libs.lib.fast_zip([
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_FAC]['RiskAttachment'].values,
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_FAC]['RiskLimit'].values,
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_FAC]['CededPercent'].values,
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_FAC]['PlacedPercent'].values
        ]))[0] + 1 + fmprofiles_df['profile_id'].max()
        fmprofiles_df = self._get_profiles(
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_FAC],
            fmprofiles_df=fmprofiles_df, attachment="row['RiskAttachment']",
            limit="row['RiskLimit']", ceded="row['CededPercent']",
            placement="row['PlacedPercent']"
        )
        # Per Risk profile IDs
        self.logger.debug(f'{oed.REINS_TYPE_PER_RISK} profiles...')
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_PER_RISK, 'level_id'] = self._get_risk_level_id()
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_PER_RISK, 'profile_id'] = pd.factorize(pd._libs.lib.fast_zip([
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_PER_RISK]['RiskAttachment'].values,
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_PER_RISK]['RiskLimit'].values,
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_PER_RISK]['CededPercent'].values
        ]))[0] + 1 + fmprofiles_df['profile_id'].max()
        fmprofiles_df = self._get_profiles(
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_PER_RISK],
            fmprofiles_df=fmprofiles_df, attachment="row['RiskAttachment']",
            limit="row['RiskLimit']", ceded="row['CededPercent']"
        )
        # Quota Share profile IDs
        self.logger.debug(f'{oed.REINS_TYPE_QUOTA_SHARE} profiles...')
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_QUOTA_SHARE, 'level_id'] = self._get_risk_level_id()
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_QUOTA_SHARE, 'profile_id'] = pd.factorize(pd._libs.lib.fast_zip([
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_QUOTA_SHARE]['RiskLimit'].values,
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_QUOTA_SHARE]['CededPercent'].values
        ]))[0] + 1 + fmprofiles_df['profile_id'].max()
        fmprofiles_df = self._get_profiles(
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_QUOTA_SHARE],
            fmprofiles_df=fmprofiles_df, limit="row['RiskLimit']",
            ceded="row['CededPercent']"
        )
        # Surplus Share profile IDs
        self.logger.debug(f'{oed.REINS_TYPE_SURPLUS_SHARE} profiles...')
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_SURPLUS_SHARE, 'level_id'] = self._get_risk_level_id()
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_SURPLUS_SHARE, 'profile_id'] = pd.factorize(pd._libs.lib.fast_zip([
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_SURPLUS_SHARE]['RiskAttachment'].values,
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_SURPLUS_SHARE]['RiskLimit'].values,
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_SURPLUS_SHARE]['CededPercent_scope'].values
        ]))[0] + 1 + fmprofiles_df['profile_id'].max()
        fmprofiles_df = self._get_profiles(
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_SURPLUS_SHARE],
            fmprofiles_df=fmprofiles_df, attachment="row['RiskAttachment']",
            limit="row['RiskLimit']", ceded="row['CededPercent_scope']"
        )
        # Cat XL profile IDs = pass through profile ID
        self.logger.debug(f'{oed.REINS_TYPE_CAT_XL} profiles...')
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_CAT_XL, 'level_id'] = self._get_risk_level_id()
        ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_CAT_XL, 'profile_id'] = passthroughprofile_id

        ri_df['profile_id'] = ri_df['profile_id'].astype('int64')

        fields = self._get_reins_type_fields(all_fields=True)
        valid_fields = self._get_valid_fields(fields)
        for field, valid_field in zip(fields, valid_fields):
            ri_df[valid_field] = ri_df.apply(lambda row: self._is_valid_id(row[field]), axis=1)
        if self.logger:
            self._log_dataframe({'ri_df': ri_df})

        # Assign risk level and filter level profile IDs
        self.logger.debug('Assigning risk level and filter level profile IDs:')
        # Facultative profile IDs
        self.logger.debug(f'{oed.REINS_TYPE_FAC} profiles...')
        layer_id_set = set(
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_FAC]['layer_id']
        )
        # Risk level
        # Note that Facultative profiles scope much match the filter exactly
        profile_map_df.loc[
            profile_map_df['layer_id'].isin(layer_id_set), 'profile_id'
        ] = self._get_risk_level_profile_ids(
            df_=profile_map_df[profile_map_df['layer_id'].isin(layer_id_set)],
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_FAC], exact=True
        )
        # Filter level
        profile_map_df.loc[
            (profile_map_df['layer_id'].isin(layer_id_set)) & (profile_map_df['level_id'] == self._get_filter_level_id()),
            'profile_id'
        ] = passthroughprofile_id

        # Per Risk profile IDs
        self.logger.debug(f'{oed.REINS_TYPE_PER_RISK} profiles...')
        layer_id_set = set(
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_PER_RISK]['layer_id']
        )
        # Risk level
        profile_map_df.loc[
            profile_map_df['layer_id'].isin(layer_id_set), 'profile_id'
        ] = self._get_risk_level_profile_ids(
            df_=profile_map_df[profile_map_df['layer_id'].isin(layer_id_set)],
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_PER_RISK]
        )
        # Filter level
        filter_idx = self._get_filter_level_profile_ids(
            df_=profile_map_df[(profile_map_df['layer_id'].isin(layer_id_set)) & (profile_map_df['level_id'] == self._get_filter_level_id())],
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_PER_RISK][fields + valid_fields + ['layer_id']]
        )
        profile_map_df.loc[filter_idx, 'profile_id'] = passthroughprofile_id

        # Quota Share profile IDs
        self.logger.debug(f'{oed.REINS_TYPE_QUOTA_SHARE} profiles...')
        layer_id_set = set(
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_QUOTA_SHARE]['layer_id']
        )
        # Risk level
        profile_map_df.loc[
            profile_map_df['layer_id'].isin(layer_id_set), 'profile_id'
        ] = self._get_risk_level_profile_ids(
            df_=profile_map_df[profile_map_df['layer_id'].isin(layer_id_set)],
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_QUOTA_SHARE]
        )
        # Filter level
        filter_idx = self._get_filter_level_profile_ids(
            df_=profile_map_df[(profile_map_df['layer_id'].isin(layer_id_set)) & (profile_map_df['level_id'] == self._get_filter_level_id())],
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_QUOTA_SHARE][fields + valid_fields + ['layer_id']]
        )
        profile_map_df.loc[filter_idx, 'profile_id'] = passthroughprofile_id

        # Surplus Share profile IDs
        self.logger.debug(f'{oed.REINS_TYPE_SURPLUS_SHARE} profiles...')
        layer_id_set = set(
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_SURPLUS_SHARE]['layer_id']
        )
        # Risk level
        # Note that Surplus Share profiles scope much match the filter exactly
        profile_map_df.loc[
            profile_map_df['layer_id'].isin(layer_id_set), 'profile_id'
        ] = self._get_risk_level_profile_ids(
            df_=profile_map_df[profile_map_df['layer_id'].isin(layer_id_set)],
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_SURPLUS_SHARE],
            exact=True
        )
        # Filter level
        profile_map_df.loc[
            (profile_map_df['layer_id'].isin(layer_id_set)) & (profile_map_df['level_id'] == self._get_filter_level_id()),
            'profile_id'
        ] = passthroughprofile_id

        # Cat XL profile IDs
        self.logger.debug(f'{oed.REINS_TYPE_CAT_XL} profiles...')
        layer_id_set = set(
            ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_CAT_XL]['layer_id']
        )
        # Risk level
        profile_map_df.loc[
            profile_map_df['layer_id'].isin(layer_id_set), 'profile_id'
        ] = self._get_risk_level_profile_ids(
            df_=profile_map_df[profile_map_df['layer_id'].isin(layer_id_set)],
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_CAT_XL],
        )
        # Filter level
        filter_idx = self._get_filter_level_profile_ids(
            df_=profile_map_df[(profile_map_df['layer_id'].isin(layer_id_set)) & (profile_map_df['level_id'] == self._get_filter_level_id())],
            ri_df_=ri_df[ri_df['ReinsType'] == oed.REINS_TYPE_CAT_XL][fields + valid_fields + ['layer_id']]
        )
        profile_map_df.loc[filter_idx, 'profile_id'] = passthroughprofile_id

        # OccLimit / Placed Percent
        occlimit_df = self.ri_info_df.copy()
        occlimit_df = occlimit_df.merge(ri_df[occlimit_df.columns.to_list() + ['layer_id']], how='left').drop_duplicates()
        occlimit_df = occlimit_df[occlimit_df['ReinsType'] != oed.REINS_TYPE_FAC]
        occlimit_df['level_id'] = self._get_risk_level_id() + 1
        occlimit_df = occlimit_df.reset_index(drop=True)
        occlimit_df['profile_id'] = occlimit_df.index + 1 + ri_df['profile_id'].max()
        if self.logger:
            self._log_dataframe({'occlimit_df': occlimit_df})

        # Assign programme node level profile IDs
        self.logger.debug('Assigning programme node level profile IDs')
        # Per Risk, Quota Share & Surplus Share profile IDs
        fmprofiles_df = self._get_occlimit_profiles(
            occlimit_df_=occlimit_df[occlimit_df['ReinsType'] != oed.REINS_TYPE_CAT_XL],
            fmprofiles_df=fmprofiles_df, limit="row['OccLimit']",
            placement="row['PlacedPercent']"
        )
        # Cat XL profile IDs
        fmprofiles_df = self._get_profiles(
            ri_df_=occlimit_df[occlimit_df['ReinsType'] == oed.REINS_TYPE_CAT_XL],
            fmprofiles_df=fmprofiles_df, attachment="row['OccAttachment']",
            limit="row['OccLimit']", ceded="row['CededPercent']",
            placement="row['PlacedPercent']"
        )

        # Programme level profile IDs
        profile_map_df.loc[
            profile_map_df['level_id'] == 4, 'profile_id'
        ] = self._match_exact(
            profile_map_df[profile_map_df['level_id'] == 4],
            occlimit_df[['layer_id', 'level_id', 'profile_id']]
        )

        profile_map_df['profile_id'] = profile_map_df['profile_id'].astype('int64')
        if self.logger:
            self._log_dataframe({'profile_map_df': profile_map_df})

        # Write out Oasis structure
        self.fmprogrammes_df = xref_df[xref_df['agg_id_to'] != 0][['agg_id', 'level_id', 'agg_id_to']].reset_index(drop=True)
        self.fmprogrammes_df.columns = ['from_agg_id', 'level_id', 'to_agg_id']
        self.fmprofiles_df = fmprofiles_df.sort_values(by='profile_id').reset_index(drop=True)
        self.fm_policytcs_df = profile_map_df[profile_map_df['level_id'] > 1][['layer_id', 'level_id', 'agg_id', 'profile_id']].reset_index(drop=True)
        self.fm_policytcs_df['level_id'] = self.fm_policytcs_df['level_id'] - 1

        if self.logger:
            self._log_dataframe({
                'fm_programmes_df': self.fmprogrammes_df,
                'fmprofiles_df': self.fmprofiles_df,
                'fm_policytcs_df': self.fm_policytcs_df
            })

    def write_oasis_files(self, directory=None):
        '''
        Write out the generated data to Oasis input file format.
        '''

        if directory is None:
            directory = "direct"
        if os.path.exists(directory):
            shutil.rmtree(directory)
        os.makedirs(directory)

        self.coverages_df.to_csv(
            os.path.join(directory, "coverages.csv"), index=False)
        self.items_df.to_csv(
            os.path.join(directory, "items.csv"), index=False)
        self.fmprogrammes_df.to_csv(
            os.path.join(directory, "fm_programme.csv"), index=False)
        self.fmprofiles_df.to_csv(
            os.path.join(directory, "fm_profile.csv"), index=False)
        self.fm_policytcs_df.to_csv(
            os.path.join(directory, "fm_policytc.csv"), index=False)
        self.fmsummaryxref_df.to_csv(
            os.path.join(directory, "fmsummaryxref.csv"), index=False)
        self.gulsummaryxref_df.to_csv(
            os.path.join(directory, "gulsummaryxref.csv"), index=False)



import numba as nb
import numpy as np

AggregateVulnerability = nb.from_dtype(np.dtype([('aggregate_vulnerability_id', np.int32),
                                                 ('vulnerability_id', np.int32),]))

VulnerabilityWeight = nb.from_dtype(np.dtype([('areaperil_id', np.int32),
                                              ('vulnerability_id', np.int32),
                                              ('weight', np.int32)]))



"""
This file contains general-purpose utilities.
"""
import numpy as np


def assert_allclose(x, y, rtol=1e-10, atol=1e-8, x_name="x", y_name="y"):
    """
    Drop in replacement for `numpy.testing.assert_allclose` that also shows
    the nonmatching elements in a nice human-readable format.

    Args:
        x (np.array or scalar): first input to compare.
        y (np.array or scalar): second input to compare.
        rtol (float, optional): relative tolreance. Defaults to 1e-10.
        atol (float, optional): absolute tolerance. Defaults to 1e-8.
        x_name (str, optional): header to print for x if x and y do not match. Defaults to "x".
        y_name (str, optional): header to print for y if x and y do not match. Defaults to "y".

    Raises:
        AssertionError: if x and y shapes do not match.
        AssertionError: if x and y data do not match.

    """
    if np.isscalar(x) and np.isscalar(y) == 1:
        return np.testing.assert_allclose(x, y, rtol=rtol, atol=atol)

    if x.shape != y.shape:
        raise AssertionError("Shape mismatch: %s vs %s" % (str(x.shape), str(y.shape)))

    d = ~np.isclose(x, y, rtol, atol)

    if np.any(d):
        miss = np.where(d)[0]
        msg = f"Mismatch of {len(miss):d} elements ({len(miss) / x.size * 100:g} %) at the level of rtol={rtol:g}, atol={atol:g},\n" \
            f"{repr(miss)}\n" \
            f"x: {x_name}\n{str(x[d])}\n\n" \
            f"y: {y_name}\n{str(y[d])}"\

        raise AssertionError(msg)



__all__ = [
    'set_calc_rule_ids',
    'get_calc_rule_ids',
    'get_grouped_fm_profile_by_level_and_term_group',
    'get_grouped_fm_terms_by_level_and_term_group',
    'get_oed_hierarchy',
    'get_il_input_items',
    'get_policytc_ids',
    'get_step_calc_rule_ids',
    'get_step_policytc_ids',
    'write_il_input_files',
    'write_fm_policytc_file',
    'write_fm_profile_file',
    'write_fm_programme_file',
    'write_fm_xref_file'
]

import copy
import itertools
import os
import sys
import warnings

import pandas as pd
import numpy as np

from ast import literal_eval

from ..utils.calc_rules import (
    get_calc_rules,
    get_step_calc_rules
)
from ..utils.coverages import SUPPORTED_COVERAGE_TYPES
from ..utils.data import (
    factorize_array,
    factorize_ndarray,
    fast_zip_arrays,
    merge_check,
    merge_dataframes,
    set_dataframe_column_dtypes,
    fill_na_with_categoricals,
)
from ..utils.defaults import (
    assign_defaults_to_il_inputs,
    get_default_accounts_profile,
    get_default_exposure_profile,
    get_default_fm_aggregation_profile,
    OASIS_FILES_PREFIXES,
)
from ..utils.exceptions import OasisException
from ..utils.fm import (
    DEDUCTIBLE_AND_LIMIT_TYPES,
    SUPPORTED_FM_LEVELS,
    STEP_TRIGGER_TYPES,
    COVERAGE_AGGREGATION_METHODS,
    CALCRULE_ASSIGNMENT_METHODS,
)
from ..utils.log import oasis_log
from ..utils.path import as_path
from ..utils.profiles import (
    get_fm_terms_oed_columns,
    get_grouped_fm_profile_by_level_and_term_group,
    get_grouped_fm_terms_by_level_and_term_group,
    get_oed_hierarchy,
    get_default_step_policies_profile,
)
from .summaries import get_useful_summary_cols, get_xref_df

pd.options.mode.chained_assignment = None
warnings.simplefilter(action='ignore', category=FutureWarning)

# Define a list of all supported OED coverage types in the exposure
supp_cov_types = [v['id'] for v in SUPPORTED_COVERAGE_TYPES.values()]

step_profile_cols = [
    'policytc_id', 'calcrule_id',
    'deductible1', 'deductible2', 'deductible3', 'attachment1',
    'limit1', 'share1', 'share2', 'share3', 'step_id',
    'trigger_start', 'trigger_end', 'payout_start', 'payout_end',
    'limit2', 'scale1', 'scale2'
]

profile_cols = ['policytc_id', 'calcrule_id', 'deductible1', 'deductible2', 'deductible3', 'attachment1', 'limit1',
                'share1', 'share2', 'share3']
profile_cols_map = {
    'deductible': 'deductible1',
    'deductible_min': 'deductible2',
    'deductible_max': 'deductible3',
    'attachment': 'attachment1',
    'limit': 'limit1',
    'share': 'share1'
}


def set_calc_rule_ids(
        il_inputs_calc_rules_df,
        terms,
        terms_indicators,
        types_and_codes,
        types,
        policy_layer=False
):
    """
    Lookup and assign calc. rule IDs

    :param il_inputs_df: IL inputs items dataframe
    :type il_inputs_df: pandas.DataFrame

    :param terms: list of relevant terms
    :type terms: list

    :param terms_indicators: list indicating whether relevant terms > 0
    :type terms_indicators: list

    :param types_and_codes: list of types and codes if applicable
    :type types_and_codes: list

    :param types: list of types if applicable
    :type types: list or NoneType

    :param policy_layer: flag to indicate whether policy layer present
    :type policy_layer: bool

    :return: Numpy array of calc. rule IDs
    :type: numpy.ndarray
    """

    calc_rules = get_calc_rules(policy_layer).drop(['desc'], axis=1)
    try:
        calc_rules['id_key'] = calc_rules['id_key'].apply(literal_eval)
    except ValueError as e:
        raise OasisException("Exception raised in 'set_calc_rule_ids'", e)

    # Percentage type assigned same calc. rule as flat type so set both to flat
    # type to reduce permutations
    if types:
        il_inputs_calc_rules_df.loc[:, types] = np.where(
            il_inputs_calc_rules_df[types] == DEDUCTIBLE_AND_LIMIT_TYPES['pctiv']['id'],
            DEDUCTIBLE_AND_LIMIT_TYPES['flat']['id'],
            il_inputs_calc_rules_df[types]
        )
    il_inputs_calc_rules_df.loc[:, terms_indicators] = np.where(
        il_inputs_calc_rules_df[terms] > 0, 1, 0
    )
    il_inputs_calc_rules_df['id_key'] = [
        t for t in fast_zip_arrays(*il_inputs_calc_rules_df.loc[:, terms_indicators + types_and_codes].transpose().values)
    ]
    il_inputs_calc_rules_df = merge_dataframes(
        il_inputs_calc_rules_df, calc_rules, how='left', on='id_key', drop_duplicates=False
    ).fillna(0)
    il_inputs_calc_rules_df['calcrule_id'] = il_inputs_calc_rules_df['calcrule_id'].astype('uint32')
    if 0 in il_inputs_calc_rules_df.calcrule_id.unique():
        err_msg = 'Calculation Rule mapping error, non-matching keys:\n'
        no_match_keys = il_inputs_calc_rules_df.loc[il_inputs_calc_rules_df.calcrule_id == 0].id_key.unique()

        err_msg += '   {}\n'.format(tuple(terms_indicators + types_and_codes))
        for key_id in no_match_keys:
            err_msg += '   {}\n'.format(key_id)
        raise OasisException(err_msg)

    return il_inputs_calc_rules_df['calcrule_id'].values


def get_calc_rule_ids(il_inputs_df):
    """
    Returns a Numpy array of calc. rule IDs from a table of IL input items

    :param il_inputs_df: IL input items dataframe
    :type il_inputs_df: pandas.DataFrame

    :return: Numpy array of calc. rule IDs
    :rtype: numpy.ndarray
    """
    policy_layer_id = SUPPORTED_FM_LEVELS['policy layer']['id']

    # Get calc. rule IDs for all levels except policy layer
    terms = ['deductible', 'deductible_min', 'deductible_max', 'limit']
    terms_indicators = ['{}_gt_0'.format(t) for t in terms]
    types_and_codes = ['ded_type', 'ded_code', 'lim_type', 'lim_code']
    types = ['ded_type', 'lim_type']
    calc_mapping_cols = ['item_id'] + terms + terms_indicators + types_and_codes + ['orig_level_id', 'calcrule_id']
    il_inputs_calc_rules_df = il_inputs_df.reindex(columns=calc_mapping_cols)
    il_inputs_df.loc[
        il_inputs_df['orig_level_id'] != policy_layer_id, 'calcrule_id'
    ] = set_calc_rule_ids(
        il_inputs_calc_rules_df[il_inputs_calc_rules_df['orig_level_id'] != policy_layer_id],
        terms, terms_indicators, types_and_codes, types
    )

    # Get calc. rule IDs for policy layer level
    terms = ['limit', 'share', 'attachment']
    terms_indicators = ['{}_gt_0'.format(t) for t in terms]
    types_and_codes = []
    types = None
    calc_mapping_cols = ['item_id'] + terms + terms_indicators + ['orig_level_id', 'calcrule_id']
    il_inputs_calc_rules_df = il_inputs_df.reindex(columns=calc_mapping_cols)
    il_inputs_df.loc[
        il_inputs_df['orig_level_id'] == policy_layer_id, 'calcrule_id'
    ] = set_calc_rule_ids(
        il_inputs_calc_rules_df[il_inputs_calc_rules_df['orig_level_id'] == policy_layer_id],
        terms, terms_indicators, types_and_codes, types, policy_layer=True
    )

    return il_inputs_df['calcrule_id'].values


def get_step_calc_rule_ids(il_inputs_df):
    """
    Returns a Numpy array of calc. rule IDs from a table of IL input items that
    include step policies

    :param il_inputs_df: IL input items dataframe
    :type il_inputs_df: pandas.DataFrame

    :return: Numpy array of calc. rule IDs
    :rtype: numpy.ndarray
    """
    calc_rules_step = get_step_calc_rules().drop(['desc'], axis=1)
    try:
        calc_rules_step['id_key'] = calc_rules_step['id_key'].apply(literal_eval)
    except ValueError as e:
        raise OasisException("Exception raised in 'get_step_calc_rule_ids'", e)

    terms = ['deductible1', 'payout_start', 'payout_end', 'limit1', 'limit2']
    terms_indicators = ['{}_gt_0'.format(t) for t in terms]
    types = ['trigger_type', 'payout_type']

    cols = ['orig_level_id',
            'item_id', 'level_id', 'StepTriggerType', 'assign_step_calcrule',
            'coverage_type_id'
            ]

    calc_mapping_cols = cols + terms + terms_indicators + types + ['calcrule_id']
    il_inputs_calc_rules_df = il_inputs_df.reindex(columns=calc_mapping_cols)

    il_inputs_calc_rules_df.loc[:, terms_indicators] = np.where(il_inputs_calc_rules_df[terms] > 0, 1, 0)
    il_inputs_calc_rules_df[types] = il_inputs_calc_rules_df[types].fillna(0).astype('uint8')
    il_inputs_calc_rules_df['id_key'] = [t for t in fast_zip_arrays(*il_inputs_calc_rules_df.loc[:, terms_indicators + types].transpose().values)]
    il_inputs_calc_rules_df = merge_dataframes(il_inputs_calc_rules_df, calc_rules_step, how='left', on='id_key', drop_duplicates=False).fillna(0)

    # Assign passthrough calcrule ID 100 to first level
    il_inputs_calc_rules_df['calcrule_id'] = il_inputs_calc_rules_df['calcrule_id'].astype('uint32')
    if 0 in il_inputs_calc_rules_df.calcrule_id.unique():
        err_msg = 'Calculation Rule mapping error, non-matching keys:\n'
        no_match_keys = il_inputs_calc_rules_df.loc[il_inputs_calc_rules_df.calcrule_id == 0].id_key.unique()

        err_msg += '   {}\n'.format(tuple(terms_indicators + types))
        for key_id in no_match_keys:
            err_msg += '   {}\n'.format(key_id)
        raise OasisException(err_msg)

    return il_inputs_calc_rules_df['calcrule_id'].values


def get_policytc_ids(il_inputs_df):
    """
    Returns a Numpy array of policy TC IDs from a table of IL input items

    :param il_inputs_df: IL input items dataframe
    :type il_inputs_df: pandas.DataFrame

    :return: Numpy array of policy TC IDs
    :rtype: numpy.ndarray
    """
    policytc_cols = ['calcrule_id', 'limit', 'deductible', 'deductible_min', 'deductible_max', 'attachment', 'share']
    return factorize_ndarray(il_inputs_df.loc[:, policytc_cols].values, col_idxs=range(len(policytc_cols)))[0]


def get_step_policytc_ids(
        il_inputs_df,
        offset=0,
        idx_cols=[]
):
    """
    Returns a Numpy array of policy TC IDs from a table of IL input items that
    include step policies

    :param il_inputs_df: IL input items dataframe
    :type il_inputs_df: pandas.DataFrame

    :param step_trigger_type_cols: column names used to determine values for
    terms indicators and types
    :type step_trigger_type_cols: list

    :return: Numpy array of policy TC IDs
    :rtype: numpy.ndarray
    """
    fm_policytc_cols = [col for col in idx_cols + ['layer_id', 'level_id', 'agg_id', 'coverage_id', 'assign_step_calcrule'] + step_profile_cols
                        if col in il_inputs_df.columns]
    fm_policytc_df = il_inputs_df[fm_policytc_cols]
    fm_policytc_df['policytc_id'] = factorize_ndarray(fm_policytc_df.loc[:, ['layer_id', 'level_id', 'agg_id']].values, col_idxs=range(3))[0]
    fm_policytc_df['pol_id'] = factorize_ndarray(fm_policytc_df.loc[:, idx_cols + ['coverage_id']].values, col_idxs=range(len(idx_cols) + 1))[0]

    step_calcrule_policytc_agg = pd.DataFrame(
        fm_policytc_df[fm_policytc_df['assign_step_calcrule'] == True]['policytc_id'].to_list(),
        index=fm_policytc_df[fm_policytc_df['assign_step_calcrule'] == True]['pol_id']
    ).groupby('pol_id').aggregate(list).to_dict()[0]

    fm_policytc_df.loc[
        fm_policytc_df['assign_step_calcrule'] == True, 'policytc_id'
    ] = fm_policytc_df.loc[fm_policytc_df['assign_step_calcrule'] == True]['pol_id'].map(step_calcrule_policytc_agg)
    fm_policytc_df['policytc_id'] = fm_policytc_df['policytc_id'].apply(
        lambda x: tuple(x) if isinstance(x, list) else x
    )
    return factorize_array(fm_policytc_df['policytc_id'])[0] + offset


def __level_has_fm_terms(df, fm_column_names):
    """Check if a df has non null values for a list of column name
    :param df: dataframe
    :type df: pandas.DataFrame

    :param fm_column_names: column names used
    :type fm_column_names: list

    :return: True is any column has at least one non null value
    :rtype: bool
    """
    return df.loc[:, list(set(fm_column_names).intersection(df.columns))].any().any()


def __get_bi_tiv_col_name(profile):
    """get the name of the bi tiv column
    :param profile: profile containing all the terms
    :type profile: dict

    :return: the name of the bi tiv column
    :rtype: str
    """
    cov_level_id = SUPPORTED_FM_LEVELS['site coverage']['id']
    bi_cov_id = SUPPORTED_COVERAGE_TYPES['bi']['id']
    return profile[cov_level_id][bi_cov_id]['tiv']['ProfileElementName']


def get_programme_ids(il_inputs_df, level):
    """
    Returns a Serise of Agg ids by level for the FM programme file generation

    :param il_inputs_df: IL input items dataframe
    :type il_inputs_df: pandas.DataFrame

    :param level: fm_programme level number
    :type  level: int
    """
    return il_inputs_df[il_inputs_df['level_id'] == level][['agg_id', 'coverage_id']].drop_duplicates(
        subset=['agg_id', 'coverage_id'], keep="first"
    ).agg_id.reset_index(drop=True)


def __merge_exposure_and_gul(exposure_df, gul_inputs_df, fm_terms, profile, oed_hierarchy):
    """merge the exposure df and the gul df

    :param exposure_df: exposure dataframe
    :type exposure_df: pandas.DataFrame

    :param gul_inputs_df: gul dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param fm_terms: fm terms to use
    :type fm_terms: dict

    :param profile: Source profile
    :type profile: dict

    :param oed_hierarchy: oed_hierarchy
    :type oed_hierarchy: dict

    :return: merge dataframe between exposure_df and gul_inputs_df
    :rtype: pandas.DataFrame


    """
    cond_tag = oed_hierarchy['condtag']['ProfileElementName']

    # get usefull term columns from exposure_df
    site_pd_and_site_all_term_cols = get_fm_terms_oed_columns(fm_terms, levels=['site pd', 'site all'])
    exposure_usefull_fm_cols = list(set(site_pd_and_site_all_term_cols).intersection(exposure_df.columns))

    # The coverage FM level (site coverage, #1) ID
    cov_level_id = SUPPORTED_FM_LEVELS['site coverage']['id']
    # Get the TIV column names and corresponding coverage types
    tiv_terms = {
        v['tiv']['CoverageTypeID']: v['tiv']['ProfileElementName']
        for v in profile[cov_level_id].values()
    }

    # Calculate sum of TIV columns
    exposure_df['tiv_sum'] = exposure_df[tiv_terms.values()].sum(axis=1)

    # set default cond_tag
    if cond_tag not in exposure_df.columns:
        exposure_df[cond_tag] = '0'
    else:
        fill_na_with_categoricals(exposure_df, {cond_tag: '0'})
        exposure_df.loc[exposure_df[cond_tag] == '', cond_tag] = '0'

    # Identify BI TIV column
    bi_tiv_col = __get_bi_tiv_col_name(profile)

    # merge exposure_df and gul_inputs_df #####
    return merge_dataframes(
        exposure_df.loc[:, exposure_usefull_fm_cols + [
            'loc_id', 'tiv_sum', bi_tiv_col, cond_tag
        ]],
        gul_inputs_df,
        join_on='loc_id',
        how='inner'
    ).rename(columns={'item_id': 'gul_input_id'})


def __merge_gul_and_account(gul_inputs_df, accounts_df, fm_terms, oed_hierarchy):
    """prepare gul and account df and merge them based on [portfolio_num, acc_num, cond_tag]"""

    acc_num = oed_hierarchy['accnum']['ProfileElementName']
    portfolio_num = oed_hierarchy['portnum']['ProfileElementName']
    policy_num = oed_hierarchy['polnum']['ProfileElementName']
    cond_tag = oed_hierarchy['condtag']['ProfileElementName']
    cond_num = oed_hierarchy['condnum']['ProfileElementName']
    cond_class = oed_hierarchy['condclass']['ProfileElementName']
    loc_num = oed_hierarchy['locnum']['ProfileElementName']

    ###### prepare accounts_df #####
    # add default cond_tag if needed
    if cond_tag in gul_inputs_df:
        def add_default_value(df, col, value):
            if col not in df:
                df[col] = value
                df[col] = df[col].astype('category')
            else:
                if value not in df[col].cat.categories:
                    df[col] = df[col].cat.add_categories(value)
                df[col] = df[col].fillna(value)

        add_default_value(accounts_df, cond_tag, '0')
        add_default_value(accounts_df, cond_num, '')

    if cond_tag in accounts_df.columns:
        # Oed schema allows policy levels to contain location with and without specific cond class
        # for a specific layer, a location without cond class for this layer is pass through
        # if the cond class for other location is 0 and excluded if cond class is 1

        # first we determine if the layer has the pass through or the exclusion rule and store the result in polcondclass
        if cond_class in accounts_df.columns:  # cond_class is use to specify if a condition exclude other location (1) or not (0)
            # policy level cond class is set to 1 if some location are excluded
            accounts_df['PolCondClass'] = (accounts_df.groupby([portfolio_num, acc_num, policy_num, 'layer_id'], sort=False)
                                           [cond_class]
                                           .transform(max))
        else:
            # otherwise they are set to 0 (no exclusion)
            accounts_df[['PolCondClass', cond_class]] = 0

        if 'CondPriority' not in accounts_df.columns:
            accounts_df['CondPriority'] = 1
        accounts_df['CondPriority'].fillna(1, inplace=True)

        # create a df all_cond_policy containing all the cond_tag for each policies
        policy_df = accounts_df.drop_duplicates(subset=[portfolio_num, acc_num, policy_num, 'layer_id']).drop(columns=[cond_tag, 'CondPriority'])
        cond_df = accounts_df[[portfolio_num, acc_num, cond_tag, 'CondPriority']].drop_duplicates()
        all_cond_policy = (pd.merge(policy_df, cond_df, on=[portfolio_num, acc_num])
                           .drop(columns=cond_class)
                           .rename(columns={'PolCondClass': cond_class}))

        # get which cond tag are not specified in which layer
        missing_cond_policy_df = pd.merge(accounts_df[[portfolio_num, acc_num, policy_num, 'layer_id', cond_tag]],
                                          all_cond_policy, how='right', indicator=True)
        missing_cond_policy_df = missing_cond_policy_df[missing_cond_policy_df['_merge'] == 'right_only'].drop(columns='_merge')

        # finally, we can create default cond that will be applied to the location that have no cond in some layer
        # depending on the cond_class, the default cond will be passing through or excluding the loss
        null_cond = accounts_df[accounts_df[cond_tag] != '0']
        null_cond[cond_tag] = '0'
        null_cond = pd.concat([null_cond, missing_cond_policy_df])
        null_cond[cond_num] = ''
        level_id = SUPPORTED_FM_LEVELS['cond all']['id']
        level_term_cols = get_fm_terms_oed_columns(fm_terms, level_ids=[level_id])
        null_cond[level_term_cols] = 0
        null_cond.drop_duplicates(subset=[portfolio_num, acc_num, cond_tag, 'layer_id'], inplace=True)
        if cond_class in null_cond.columns:
            filter_cond = (null_cond[cond_class] == 1)
            if filter_cond.any():
                null_cond.loc[filter_cond, cond_num] = 'FullFilter'
                null_cond.loc[filter_cond, 'CondDed6All'] = 1
                null_cond.loc[filter_cond, 'CondDedType6All'] = 1

        accounts_df = pd.concat([accounts_df, null_cond])
        merge_col = [portfolio_num, acc_num, cond_tag]
    else:
        merge_col = [portfolio_num, acc_num]

    ##### merge accounts_df and gul_inputs_df #####
    # check for empty intersection between dfs
    merge_check(
        gul_inputs_df[merge_col],
        accounts_df[merge_col],
        on=merge_col
    )

    # Construct a basic IL inputs frame by merging the combined exposure +
    # GUL inputs frame above, with the accounts frame, on portfolio no.,
    # account no. and condition no. (by default items in the GUL inputs frame
    # are set with a condition no. of 0)
    column_base_il_df = merge_dataframes(
        gul_inputs_df,
        accounts_df,
        on=merge_col,
        how='left',
        drop_duplicates=True
    )
    missing_account_row = column_base_il_df.loc[column_base_il_df['layer_id'].isna()]
    if not missing_account_row.empty:
        raise OasisException("locations have policies, accounts combination not present in the account file \n" +
                             missing_account_row[[loc_num] + merge_col].drop_duplicates().to_string(index=False))

    # If the merge is empty raise an exception - this will happen usually
    # if there are no common acc. numbers between the GUL input items and
    # the accounts listed in the accounts file
    if column_base_il_df.empty:
        raise OasisException(
            'Inner merge of the GUL inputs + exposure file dataframe '
            'and the accounts file dataframe  on acc. number '
            'is empty - '
            'please check that the acc. number columns in the exposure '
            'and accounts files respectively have a non-empty '
            'intersection'
        )
    return column_base_il_df


def __update_level(group, min_level, groups):
    """
    update the sub levels of all the parents of a group
    the sublevel is increased by 1 to match the depth of the childs of the parent group
    """

    while True:
        if group['level'] < min_level:
            group['level'] = min_level
            if group.get('parent') is not None:
                group = groups[group['parent']]
                min_level += 1
            else:

                return min_level
        else:
            return min_level


def __extract_level_location_to_agg_cond_dict(account_groups, base_level, max_level):
    """
    from the groups of account_groups, create a dictionary that will give the mapping between
    (ptf, account, level_id, layer_id, location) => (agg_id, condition)
    """
    level_location_to_agg_cond_dict = {}
    level_grouping = {}
    agg_dict = {level_id: 1 for level_id in range(base_level, max_level + 1)}

    def set_agg_cond(group_id, group, level_id, layer_id, condition):
        level_group_key = group_id + (level_id,)
        if level_group_key not in level_grouping:
            level_grouping[level_group_key] = agg_dict[level_id]
            agg_dict[level_id] += 1
        for loc_key in group['locations']:
            location = loc_key[-1]
            level_location_to_agg_cond_dict[group_id + (level_id, layer_id, location)] = (level_grouping[level_group_key], condition)

    for main_key, groups in account_groups.items():
        # first we iter through groups to create agg
        no_parent_not_top_groups = dict(groups)
        for group_id, group in groups.items():
            for layer_id, cond_num in group['layers'].items():
                for level_id in range(base_level, group['level']):
                    set_agg_cond(group_id, group, level_id, layer_id, '')

                if group['level'] == max_level:
                    no_parent_not_top_groups.pop(group_id, None)
                set_agg_cond(group_id, group, group['level'], layer_id, cond_num)

                for child_group_id in group.get('childs', []):
                    no_parent_not_top_groups.pop(child_group_id, None)
                    set_agg_cond(group_id, groups[child_group_id], group['level'], layer_id, cond_num)

        for group_id in no_parent_not_top_groups:
            for level_id in range(group['level'] + 1, max_level + 1):
                group = groups[group_id]
                for layer_id in group['layers']:
                    set_agg_cond(group_id, group, level_id, layer_id, '')

    return level_location_to_agg_cond_dict


def __get_cond_grouping_hierarchy(column_base_il_df, main_key, cond_tag, cond_num, loc_num, base_level):
    """create group of locations based on the condition tag, condition number and condition hierarchy found in column_base_il_df

    """

    def attach_cond(child_group, parent_group, cond_to_group):
        child_group['parent'] = parent_group['tag'][0]
        parent_group.setdefault('childs', {})[child_group['tag'][0]] = None

        cur_parent_cond = parent_group['tag'][0]
        min_level = child_group['level'] + 1
        max_loop = column_base_il_df['CondPriority'].max() + 1
        breaker = 0
        while cur_parent_cond:
            cur_parent_group = cond_to_group[cur_parent_cond]
            cur_parent_group['level'] = max(min_level, cur_parent_group['level'])
            min_level = cur_parent_group['level'] + 1
            for loc in child_group['needed_loc']:
                if loc not in cur_parent_group['needed_loc']:
                    cur_parent_group['needed_loc'][loc] = True
            cur_parent_cond = cur_parent_group.get('parent')

            if breaker > max_loop:
                raise RecursionError(f'Issue with condtag {cur_parent_cond} stuck looping through hierarchy')
            else:
                breaker += 1

    # determine the number of layers for each main_key
    main_key_layers = {}
    for rec in column_base_il_df[main_key + ['layer_id']].to_dict(orient="records"):
        main_key_layers.setdefault(tuple(rec[name] for name in main_key), set()).add(rec['layer_id'])

    loc_to_cond = {}
    account_groups = {}
    for rec in column_base_il_df[main_key + [cond_tag, cond_num, loc_num, 'layer_id', 'CondPriority']].drop_duplicates().to_dict(orient="records"):
        group_key = tuple(rec[name] for name in main_key)
        groups = account_groups.setdefault(group_key, {})
        cond_key = tuple(rec[name] for name in main_key + [cond_tag])
        loc_key = tuple(rec[name] for name in main_key + [loc_num])

        parent_cond = None
        child_cond = None

        # if location already exist we determine where the record condition is inside the location parent groups
        if loc_key in loc_to_cond:
            cur_cond = loc_to_cond[loc_key]
            while cur_cond:
                cur_group = groups[cur_cond]
                cur_condpriority = cur_group['tag'][1]
                if rec['CondPriority'] == cur_condpriority:
                    if cur_cond != cond_key:
                        raise OasisException(f'{loc_key} condition of the same priority and policy {cond_key} {cur_cond}')
                elif rec['CondPriority'] > cur_condpriority:
                    child_cond = cur_cond
                else:
                    parent_cond = cur_cond
                    break
                cur_cond = cur_group.get('parent')

        if cond_key not in groups:
            # new cond key we create a new group
            group = {'locations': {},
                     'tag': (cond_key, rec['CondPriority']),
                     'layers': {layer_id: '' for layer_id in main_key_layers[group_key]},
                     'needed_loc': {},
                     'level': base_level}
            groups[cond_key] = group
        else:
            # cond key already exist, if the parent of the cond_tag is not the same as the parent found within the location parents
            # we rearrange the hierarchy based on priority
            group = groups[cond_key]
            cond_parent_cond = group.get('parent')
            if cond_parent_cond and parent_cond and cond_parent_cond != parent_cond:
                cond_parent_group = groups[cond_parent_cond]
                loc_parent_group = groups[parent_cond]
                if cond_parent_group['tag'][1] == loc_parent_group['tag'][1]:
                    raise OasisException(f"condition of the same priority and policy {cond_parent_group['tag'][0]} {loc_parent_group['tag'][0]}")
                elif cond_parent_group['tag'][1] > loc_parent_group['tag'][1]:
                    attach_cond(loc_parent_group, cond_parent_group, groups)
                    cond_parent_group['childs'].pop(cond_key, None)
                else:
                    attach_cond(cond_parent_group, loc_parent_group, groups)
                    loc_parent_group['childs'].pop(child_cond, None)
                    parent_cond = group.get('parent')

        if child_cond:
            child_group = groups[child_cond]
            attach_cond(child_group, group, groups)
        else:
            group['locations'][loc_key] = None
            loc_to_cond[loc_key] = cond_key

        group['needed_loc'][loc_key] = False
        if group['layers'][rec['layer_id']]:
            if rec[cond_num] != group['layers'][rec['layer_id']]:
                raise ValueError(
                    f"account {group_key} has multiple cond_number ({rec[cond_num]}, {group['layers'][rec['layer_id']]})"
                    " for the same cond_tag and same layer_id")
        else:
            group['layers'][rec['layer_id']] = rec[cond_num]

        if parent_cond:
            parent_group = groups[parent_cond]
            parent_group['locations'].pop(loc_key, None)
            parent_group.get('childs', {}).pop(child_cond, None)
            attach_cond(group, parent_group, groups)

    missing_conditions = []
    max_level = base_level
    for account, groups in account_groups.items():
        for group_id, group in groups.items():
            if group['level'] > max_level:
                max_level = group['level']
            for loc_key, needed in group['needed_loc'].items():
                if needed:
                    missing_conditions.append(group['tag'][0] + ('location', loc_key[-1],))

    if missing_conditions:
        missing_conditions_df = pd.DataFrame(missing_conditions, columns=main_key + [cond_tag, 'type', 'missing'])
        raise OasisException(f"missing conditions for :\n{missing_conditions_df.to_string(index=False)}")

    return account_groups, max_level


def __get_level_location_to_agg_cond(column_base_il_df, oed_hierarchy, main_key):
    """create a dataframe with the computed agg id base on the condition and condition priority
    """
    loc_num = oed_hierarchy['locnum']['ProfileElementName']
    cond_num = oed_hierarchy['condnum']['ProfileElementName']
    cond_tag = oed_hierarchy['condtag']['ProfileElementName']

    base_level = 0
    account_groups, max_level = __get_cond_grouping_hierarchy(column_base_il_df, main_key, cond_tag, cond_num, loc_num, base_level)
    level_location_to_agg_cond_dict = __extract_level_location_to_agg_cond_dict(account_groups, base_level, max_level)

    level_location_to_agg_cond = pd.DataFrame.from_records(
        [level_location + agg_cond
         for level_location, agg_cond
         in level_location_to_agg_cond_dict.items()],
        columns=main_key + [cond_tag, 'level_id', 'layer_id', loc_num, 'agg_id', cond_num]
    )

    return level_location_to_agg_cond, max_level


def compute_agg_tiv(tiv_df, is_bi_coverage, bi_tiv_col, loc_num):
    """ compute the agg tiv depending on the agg_key"""
    agg_tiv_df = (tiv_df.drop_duplicates(['agg_id', loc_num], keep='first')[['agg_id', 'tiv', 'tiv_sum', 'is_bi_coverage', bi_tiv_col]]
                  .groupby('agg_id', observed=True).sum().reset_index())
    if is_bi_coverage:
        # we need to separate bi coverage from the other tiv
        agg_tiv_df.loc[agg_tiv_df['is_bi_coverage'] == False, 'agg_tiv'] = agg_tiv_df['tiv_sum'] - agg_tiv_df[bi_tiv_col]
        agg_tiv_df.loc[agg_tiv_df['is_bi_coverage'] == True, 'agg_tiv'] = agg_tiv_df[bi_tiv_col]
    else:
        agg_tiv_df['agg_tiv'] = agg_tiv_df['tiv_sum']
    return agg_tiv_df[['agg_id', 'agg_tiv']]


def __get_level_terms(column_base_il_df, column_mapper):
    """
    get the column to term dictionary base on a level column mapper (created from profile)
    """
    level_terms = {}
    for ProfileElementName, term_info in column_mapper.items():
        if (ProfileElementName in column_base_il_df.columns
                and column_base_il_df[ProfileElementName].any()):
            level_terms[ProfileElementName] = (term_info['FMTermType'].lower(), term_info.get('FMTermGroupID', 1))
    return level_terms


def __drop_duplicated_row(prev_level_df, level_df, level_terms, agg_key, sub_agg_key):
    """drop duplicated row base on the agg_key, sub_agg_key and layer_id"""
    sub_level_layer_needed = level_df['agg_id'].isin(prev_level_df.loc[prev_level_df["layer_id"] == 2]['to_agg_id'])

    level_value_count = level_df[['agg_id'] + list(level_terms.values())].drop_duplicates()['agg_id'].value_counts()
    this_level_layer_needed = level_df['agg_id'].isin(level_value_count[level_value_count > 1].index.values)
    if 'share' in level_df.columns:
        this_level_layer_needed |= ~level_df['share'].isin({0, 1})

    level_df['layer_id'] = np.where(sub_level_layer_needed | this_level_layer_needed,
                                    level_df['layer_id'],
                                    1)
    level_df.drop_duplicates(subset=['agg_id'] + sub_agg_key + ['layer_id'], inplace=True)


def __process_standard_level_df(column_base_il_df,
                                prev_level_df,
                                prev_agg_key,
                                tiv_df,
                                il_inputs_df_list,
                                level_id,
                                present_cols,
                                level_cols,
                                level_column_mapper,
                                bi_tiv_col,
                                oed_hierarchy,
                                fm_aggregation_profile,
                                fm_term_filters):
    # identify fm columns for this level
    level_terms_group = __get_level_terms(column_base_il_df, level_column_mapper[level_id])
    level_terms = {ProfileElementName: fm_term for (ProfileElementName, (fm_term, FMTermGroupID)) in level_terms_group.items()}

    if not level_terms:  # if no terms we skip the level
        return prev_level_df, prev_agg_key

    cur_level = len(il_inputs_df_list) + 2

    column_base_il_df['level_id'] = len(il_inputs_df_list) + 2
    column_base_il_df['orig_level_id'] = level_id

    column_base_il_df['has_terms'] = False
    for term in level_terms:
        column_base_il_df[term].fillna(0, inplace=True)
        column_base_il_df['has_terms'] |= column_base_il_df[term].astype(bool)
    agg_key = [v['field'] for v in fm_aggregation_profile[level_id]['FMAggKey'].values()]

    level_df_with_term = column_base_il_df[column_base_il_df['has_terms']]
    il_df_no_term = column_base_il_df[~column_base_il_df['has_terms']]

    prev_level_df_with_next_term = prev_level_df[prev_level_df['gul_input_id'].isin(level_df_with_term['gul_input_id'])]
    prev_level_df_no_next_term = prev_level_df[prev_level_df['gul_input_id'].isin(il_df_no_term['gul_input_id'])]

    # identify useful column name
    loc_num = oed_hierarchy['locnum']['ProfileElementName']
    if level_id in fm_term_filters:
        temp_df = pd.DataFrame(0, index=level_df_with_term.index, columns=sorted(set(level_terms.values()), key=str.lower))
        temp_df[f'filter_agg_{level_id}'] = -1

        for i, (ProfileElementName, (fm_term, FMTermGroupID)) in enumerate(level_terms_group.items()):
            filter_df = ((fm_term_filters[level_id](level_df_with_term, ProfileElementName))
                         & (~level_df_with_term[ProfileElementName].isna())
                         & (level_df_with_term[ProfileElementName].astype(bool)))
            if not level_df_with_term.loc[filter_df & ~(temp_df[f'filter_agg_{level_id}'].isin([-1, FMTermGroupID]))].empty:
                raise OasisException(f"multiple terms {fm_term} for level {level_id} in location:\n"
                                     f"{level_df_with_term.loc[filter_df & ~(temp_df[f'filter_agg_{level_id}'].isin([-1, FMTermGroupID])), agg_key]}")
            temp_df.loc[filter_df, f'filter_agg_{level_id}'] = FMTermGroupID
            temp_df.loc[filter_df, fm_term.lower()] = level_df_with_term.loc[filter_df, ProfileElementName]

        agg_key.append(f'filter_agg_{level_id}')
        level_df_with_term[temp_df.columns] = temp_df
        column_base_il_df.loc[column_base_il_df['has_terms'], f'filter_agg_{level_id}'] = temp_df[f'filter_agg_{level_id}']
    else:
        for ProfileElementName, fm_term in level_terms.items():
            level_df_with_term[fm_term] = level_df_with_term[ProfileElementName]
    level_df_with_term['orig_level_id'] = level_id

    sub_agg_key = [v['field'] for v in fm_aggregation_profile[level_id].get('FMSubAggKey', {}).values()
                   if v['field'] in level_df_with_term.columns]

    level_df_with_term['agg_id'] = factorize_ndarray(level_df_with_term.loc[:, agg_key].values, col_idxs=range(len(agg_key)))[0]
    level_df_with_term['prev_agg_id'] = factorize_ndarray(level_df_with_term.loc[:, prev_agg_key].values, col_idxs=range(len(prev_agg_key)))[0]

    # check rows in prev df that are this level granularity (if prev_agg_id has multiple corresponding agg_id)
    need_root_start_df = level_df_with_term.groupby("prev_agg_id").agg({"agg_id": pd.Series.nunique})
    need_root_start_df = need_root_start_df[need_root_start_df['agg_id'] > 1].index

    # create new prev df for element that need to restart from items
    root_df = level_df_with_term[((level_df_with_term['prev_agg_id'].isin(need_root_start_df)) & (level_df_with_term['layer_id'] == 1))]

    root_df['to_agg_id'] = root_df['agg_id']
    root_df['agg_id'] = -root_df['gul_input_id']
    root_df.drop_duplicates(subset='agg_id', inplace=True)
    root_df['level_id'] = cur_level - 1

    # rows with no parent points to 0
    prev_level_df_no_parent = prev_level_df_with_next_term[prev_level_df_with_next_term['gul_input_id'].isin(root_df['gul_input_id'])]
    prev_level_df_no_parent['to_agg_id'] = 0

    # select good to_agg_id for rows with parents
    prev_level_df_with_parent = prev_level_df_with_next_term[~prev_level_df_with_next_term['gul_input_id'].isin(
        root_df['gul_input_id'])].sort_values(by='gul_input_id')
    prev_level_df_with_parent = prev_level_df_with_parent.merge(
        level_df_with_term[['gul_input_id', 'agg_id']].rename(columns={'agg_id': 'to_agg_id'}))

    # row with no term are simply copied from previous level, they will take no time or space in subsequent fm computation
    cur_max_agg_id = level_df_with_term['agg_id'].max()
    prev_agg = prev_level_df_no_next_term[['agg_id', ]].drop_duplicates()
    prev_agg['to_agg_id'] = np.arange(cur_max_agg_id + 1, cur_max_agg_id + 1 + prev_agg.shape[0])
    prev_level_df_no_next_term = pd.merge(prev_level_df_no_next_term, prev_agg)
    # we can now copy previous no term previous level to be use in this level
    il_df_no_term = prev_level_df_no_next_term[present_cols + ['orig_level_id', 'to_agg_id']]
    il_df_no_term.rename(columns={'to_agg_id': 'agg_id'}, inplace=True)
    il_df_no_term['level_id'] = cur_level

    il_inputs_df_list.append(pd.concat([prev_level_df_with_parent, prev_level_df_no_parent,
                                        root_df, prev_level_df_no_next_term]).sort_values(by=['agg_id']))

    level_df_with_term.drop(columns=['prev_agg_id'])
    level_df = pd.concat([level_df_with_term, il_df_no_term])
    level_df['level_id'] = cur_level
    level_df.drop(columns=['agg_tiv'], errors='ignore', inplace=True)
    level_tiv_df = tiv_df.merge(level_df[['coverage_id', 'agg_id']])
    __drop_duplicated_row(prev_level_df_with_parent, level_df, level_terms, agg_key, sub_agg_key)

    # compute the aggregated tiv
    agg_tiv_df = compute_agg_tiv(level_tiv_df, 'is_bi_coverage' in agg_key, bi_tiv_col, loc_num)
    level_df = merge_dataframes(level_df, agg_tiv_df, on='agg_id', how='left')

    level_df = level_df[list(level_cols.union(set(level_terms.values())))]

    return level_df, agg_key


def __process_condition_level_df(column_base_il_df,
                                 prev_level_df,
                                 prev_agg_key,
                                 tiv_df,
                                 il_inputs_df_list,
                                 level_id,
                                 present_cols,
                                 level_cols,
                                 level_column_mapper,
                                 bi_tiv_col,
                                 oed_hierarchy,
                                 fm_aggregation_profile,
                                 fm_term_filters):
    # identify useful column name
    acc_num = oed_hierarchy['accnum']['ProfileElementName']
    portfolio_num = oed_hierarchy['portnum']['ProfileElementName']
    policy_num = oed_hierarchy['polnum']['ProfileElementName']
    cond_tag = oed_hierarchy['condtag']['ProfileElementName']
    cond_num = oed_hierarchy['condnum']['ProfileElementName']
    loc_num = oed_hierarchy['locnum']['ProfileElementName']

    if "cov_agg_id" in column_base_il_df.columns:
        main_key = [portfolio_num, acc_num, "cov_agg_id"]
    else:
        main_key = [portfolio_num, acc_num]

    # identify fm columns for this level
    level_terms_group = __get_level_terms(column_base_il_df, level_column_mapper[level_id])
    level_terms = {ProfileElementName: fm_term for (ProfileElementName, (fm_term, FMTermGroupID)) in level_terms_group.items()}

    if level_terms:  # if there is fm terms we create a new level and complete the previous level info
        level_location_to_agg_cond_df, cond_inter_level = __get_level_location_to_agg_cond(column_base_il_df,
                                                                                           oed_hierarchy, main_key)
        agg_key = [v['field'] for v in fm_aggregation_profile[level_id]['FMAggKey'].values()]
        sub_agg_key = [v['field'] for v in fm_aggregation_profile[level_id].get('FMSubAggKey', {}).values()
                       if v['field'] in column_base_il_df.columns]

        for inter_level in range(cond_inter_level + 1):
            level_df = column_base_il_df[list(set(present_cols).union(set(level_terms)))]
            level_df['level_id'] = len(il_inputs_df_list) + 2
            level_df['orig_level_id'] = level_id

            for ProfileElementName, fm_term in level_terms.items():
                level_df[fm_term] = level_df[ProfileElementName].fillna(0)

            this_level_location_to_agg_cond_df = level_location_to_agg_cond_df[level_location_to_agg_cond_df['level_id'] == inter_level]
            this_level_location_to_agg_cond_df.drop(columns=['level_id'], inplace=True)

            level_df = merge_dataframes(
                level_df,
                this_level_location_to_agg_cond_df,
                on=main_key + [loc_num, 'layer_id', cond_tag, ],
                how='inner',
                drop_duplicates=False
            )

            level_df.loc[level_df[cond_num] == '', list(set(level_terms.values()))] = 0

            this_level_location_to_agg_cond_df.rename(columns={'agg_id': 'to_agg_id'}, inplace=True)
            this_level_location_to_agg_cond_df.drop(columns=[cond_tag, cond_num], inplace=True)
            prev_level_df = merge_dataframes(
                prev_level_df,
                this_level_location_to_agg_cond_df,
                on=main_key + ['layer_id', loc_num],
                how='inner',
                drop_duplicates=False
            )
            level_tiv_df = tiv_df.merge(level_df[['coverage_id', 'agg_id']])
            __drop_duplicated_row(prev_level_df, level_df, level_terms, agg_key, sub_agg_key)

            # compute the aggregated tiv
            agg_tiv_df = compute_agg_tiv(level_tiv_df, 'is_bi_coverage' in agg_key, bi_tiv_col, loc_num)
            level_df = merge_dataframes(level_df, agg_tiv_df, on='agg_id', how='left')

            level_df = level_df[list(level_cols.union(set(level_terms.values())))]

            il_inputs_df_list.append(prev_level_df)
            prev_level_df = level_df
    else:
        agg_key = prev_agg_key

    return prev_level_df, agg_key


@oasis_log
def get_il_input_items(
        exposure_df,
        gul_inputs_df,
        accounts_df,
        exposure_profile=get_default_exposure_profile(),
        accounts_profile=get_default_accounts_profile(),
        fm_aggregation_profile=get_default_fm_aggregation_profile()
):
    """
    Generates and returns a Pandas dataframe of IL input items.

    :param exposure_df: Source exposure
    :type exposure_df: pandas.DataFrame

    :param gul_inputs_df: GUL input items
    :type gul_inputs_df: pandas.DataFrame

    :param accounts_df: Source accounts dataframe (optional)
    :param accounts_df: pandas.DataFrame

    :param accounts_fp: Source accounts file path (optional)
    :param accounts_fp: str

    :param exposure_profile: Source exposure profile (optional)
    :type exposure_profile: dict

    :param accounts_profile: Source accounts profile (optional)
    :type accounts_profile: dict

    :param fm_aggregation_profile: FM aggregation profile (optional)
    :param fm_aggregation_profile: dict

    :return: IL inputs dataframe
    :rtype: pandas.DataFrame

    :return Accounts dataframe
    :rtype: pandas.DataFrame
    """
    profile = get_grouped_fm_profile_by_level_and_term_group(exposure_profile, accounts_profile)

    # Get the FM aggregation profile - this describes how the IL input
    # items are to be aggregated in the various FM levels
    fm_aggregation_profile = copy.deepcopy(fm_aggregation_profile)

    if not fm_aggregation_profile:
        raise OasisException(
            'FM aggregation profile is empty - this is required to perform aggregation'
        )

    # Get the OED hierarchy terms profile - this defines the column names for loc.
    # ID, acc. ID, policy no. and portfolio no., as used in the source exposure
    # and accounts files. This is to ensure that the method never makes hard
    # coded references to the corresponding columns in the source files, as
    # that would mean that changes to these column names in the source files
    # may break the method
    oed_hierarchy = get_oed_hierarchy(exposure_profile, accounts_profile)
    acc_num = oed_hierarchy['accnum']['ProfileElementName']
    policy_num = oed_hierarchy['polnum']['ProfileElementName']
    portfolio_num = oed_hierarchy['portnum']['ProfileElementName']

    # get column name to fm term
    fm_terms = get_grouped_fm_terms_by_level_and_term_group(grouped_profile_by_level_and_term_group=profile, lowercase=False)
    gul_inputs_df = __merge_exposure_and_gul(exposure_df, gul_inputs_df, fm_terms, profile, oed_hierarchy)
    bi_tiv_col = 'BITIV'

    column_base_il_df = __merge_gul_and_account(gul_inputs_df, accounts_df, fm_terms, oed_hierarchy)

    # Profile dict are base on key that correspond to the fm term name.
    # this prevent multiple file column to point to the same fm term
    # which is necessary to have a generic logic that works with step policy
    # so we change the key to be the column and use FMTermType to store the term name
    level_column_mapper = {}
    for level_id, level_profile in profile.items():
        column_map = {}
        level_column_mapper[level_id] = column_map

        # for fm we only use term_id 1
        for term_name, term_info in itertools.chain.from_iterable(profile.items() for profile in level_profile.values()):
            new_term_info = copy.deepcopy(term_info)
            new_term_info['FMTermType'] = term_name
            column_map[term_info['ProfileElementName']] = new_term_info

    # column dependent fm term (level, dependency column name , dependency name in profile, default support ids)
    fm_term_filters = {}

    def site_pd_term_filter(level_df, ProfileElementName):
        return level_df['coverage_type_id'].isin(
            (level_column_mapper[SUPPORTED_FM_LEVELS['site pd']['id']].get(ProfileElementName) or {}).get('CoverageTypeID') or supp_cov_types)

    fm_term_filters[SUPPORTED_FM_LEVELS['site pd']['id']] = site_pd_term_filter

    def policy_coverage_term_filter(level_df, ProfileElementName):
        coverage_type_ids = (level_column_mapper[SUPPORTED_FM_LEVELS['policy coverage']['id']].get(
            ProfileElementName) or {}).get('CoverageTypeID') or supp_cov_types
        if isinstance(coverage_type_ids, int):
            return level_df['coverage_type_id'] == coverage_type_ids
        else:
            return level_df['coverage_type_id'].isin(coverage_type_ids)

    fm_term_filters[SUPPORTED_FM_LEVELS['policy coverage']['id']] = policy_coverage_term_filter

    # column_base_il_df contains for each items, the complete list of fm term necessary for each level
    # up until the top account level. We are now going to pivot it to get for each line a node with
    # agg_id, parrent_agg_id, level, layer and all the fm term interpretable as a generic policy
    useful_cols = sorted(set(['layer_id', 'orig_level_id', 'level_id', 'agg_id', 'gul_input_id', 'agg_tiv']
                             + get_useful_summary_cols(oed_hierarchy))
                         - {'policytc_id', 'item_id', 'output_id'}, key=str.lower)

    # Determine whether step policies are listed, are not full of nans and step
    # numbers are greater than zero
    step_policies_present = ('StepTriggerType' in column_base_il_df and 'StepNumber' in column_base_il_df
                             and column_base_il_df['StepTriggerType'].notnull().any()
                             and column_base_il_df[column_base_il_df['StepTriggerType'].notnull()]['StepNumber'].gt(0).any())

    # If step policies listed, keep step trigger type and columns associated
    # with those step trigger types that are present
    if step_policies_present:
        # we happend the fm step policy term to policy layer
        step_policy_level_map = level_column_mapper[SUPPORTED_FM_LEVELS['policy layer']['id']]
        for col in ['StepTriggerType', 'cov_agg_id', 'assign_step_calcrule']:
            step_policy_level_map[col] = {
                'ProfileElementName': col,
                'FMTermType': col,
            }
        for key, step_term in get_default_step_policies_profile().items():
            step_policy_level_map[step_term['Key']] = {
                'ProfileElementName': step_term['Key'],
                'FMTermType': step_term['FMProfileField'],
                'FMProfileStep': step_term.get('FMProfileStep')
            }

        def assign_cov_agg_id(row):
            try:
                cov_agg_method = STEP_TRIGGER_TYPES[row['StepTriggerType']]['coverage_aggregation_method']
                return COVERAGE_AGGREGATION_METHODS[cov_agg_method][row['coverage_type_id']]
            except KeyError:
                return 0

        column_base_il_df['cov_agg_id'] = column_base_il_df.apply(lambda row: assign_cov_agg_id(row), axis=1)

        def assign_calcrule_flag(row):
            try:
                calcrule_assign_method = STEP_TRIGGER_TYPES[row['StepTriggerType']]['calcrule_assignment_method']
                return CALCRULE_ASSIGNMENT_METHODS[calcrule_assign_method][row['cov_agg_id']]

            except KeyError:
                return False

        column_base_il_df['assign_step_calcrule'] = column_base_il_df.apply(lambda row: assign_calcrule_flag(row), axis=1)

        for level_info in list(SUPPORTED_FM_LEVELS.values())[1:]:
            fm_aggregation_profile[level_info['id']]['FMAggKey']['cov_agg_id'] = {
                "src": "FM",
                "field": "cov_agg_id",
                "name": "coverage aggregation id"
            }
        all_steps = column_base_il_df['StepTriggerType'].unique()

        def step_policy_term_filter(level_df, ProfileElementName):
            if 'FMProfileStep' not in level_column_mapper[SUPPORTED_FM_LEVELS['policy layer']['id']].get(ProfileElementName, {}):
                return pd.Series(True, index=level_df.index)
            else:
                return ((level_df['StepTriggerType'].isin(
                    (level_column_mapper[SUPPORTED_FM_LEVELS['policy layer']['id']].get(ProfileElementName) or {}).get('FMProfileStep')
                    or all_steps))
                    & (level_df['assign_step_calcrule'] > 0))

        fm_term_filters[SUPPORTED_FM_LEVELS['policy layer']['id']] = step_policy_term_filter

    agg_keys = set()
    for level_id in fm_aggregation_profile:
        agg_keys = agg_keys.union(set([v['field'] for v in fm_aggregation_profile[level_id]['FMAggKey'].values()]))

    level_cols = set(useful_cols).union(agg_keys)
    present_cols = [col for col in column_base_il_df.columns if col in set(useful_cols).union(agg_keys)]

    # get Tiv for each coverage
    tiv_df = column_base_il_df[sorted(
        set(agg_keys.union({'coverage_id', 'tiv', 'tiv_sum', bi_tiv_col, 'is_bi_coverage'})), key=str.lower)].drop_duplicates(keep='first')

    # initialization
    level_id = SUPPORTED_FM_LEVELS['site coverage']['id']
    coverage_level_term = ['deductible', 'deductible_min', 'deductible_max', 'limit', 'ded_code', 'ded_type',
                           'lim_code', 'lim_type']
    for col in coverage_level_term:
        if col not in column_base_il_df:
            column_base_il_df[col] = 0
    prev_level_df = column_base_il_df[list(set(present_cols + coverage_level_term))]
    prev_agg_key = [v['field'] for v in fm_aggregation_profile[level_id]['FMAggKey'].values()]
    prev_level_df.drop_duplicates(subset=prev_agg_key, inplace=True)
    prev_level_df['agg_id'] = prev_level_df['coverage_id']
    prev_level_df['level_id'] = 1
    prev_level_df['orig_level_id'] = level_id
    prev_level_df['layer_id'] = 1
    prev_level_df['agg_tiv'] = prev_level_df['tiv']
    prev_level_df['attachment'] = 0
    prev_level_df['share'] = 0
    il_inputs_df_list = []
    prev_agg_key = ['coverage_id']
    # create level for each SUPPORTED_FM_LEVELS
    for level, level_info in list(SUPPORTED_FM_LEVELS.items())[1:]:
        if level == 'cond all':  # special treatment for condition level
            process_level_df = __process_condition_level_df
        else:
            process_level_df = __process_standard_level_df

        prev_level_df, prev_agg_key = process_level_df(column_base_il_df,
                                                       prev_level_df,
                                                       prev_agg_key,
                                                       tiv_df,
                                                       il_inputs_df_list,
                                                       level_info['id'],
                                                       present_cols,
                                                       level_cols,
                                                       level_column_mapper,
                                                       bi_tiv_col,
                                                       oed_hierarchy,
                                                       fm_aggregation_profile,
                                                       fm_term_filters)

    # create account aggregation if necessary
    level = 'policy layer'
    level_id = SUPPORTED_FM_LEVELS[level]['id']
    agg_key = [v['field'] for v in fm_aggregation_profile[level_id]['FMAggKey'].values()]
    sub_agg_key = [v['field'] for v in fm_aggregation_profile[level_id].get('FMSubAggKey', {}).values()
                   if v['field'] in prev_level_df.columns]
    need_account_aggregation = prev_level_df[agg_key + sub_agg_key + ['layer_id']].groupby(agg_key + sub_agg_key + ['layer_id']).size().max() > 1

    if need_account_aggregation:
        level_df = column_base_il_df[list(set(present_cols))]
        level_df['orig_level_id'] = level_id
        level_df['level_id'] = len(il_inputs_df_list) + 2
        level_df['agg_id'] = factorize_ndarray(level_df.loc[:, agg_key].values, col_idxs=range(len(agg_key)))[0]
        prev_level_df['to_agg_id'] = factorize_ndarray(prev_level_df.loc[:, agg_key].values, col_idxs=range(len(agg_key)))[0]

        level_df.drop_duplicates(subset=agg_key + sub_agg_key + ['layer_id'], inplace=True)
        il_inputs_df_list.append(prev_level_df)
        prev_level_df = level_df

    prev_level_df['to_agg_id'] = 0
    il_inputs_df_list.append(prev_level_df)
    il_inputs_df = pd.concat(il_inputs_df_list)
    for col in il_inputs_df.columns:
        try:
            il_inputs_df[col].fillna(0, inplace=True)
        except Exception:
            pass

    # set top agg_id for later xref computation
    il_inputs_df['top_agg_id'] = factorize_ndarray(il_inputs_df.loc[:, agg_key].values, col_idxs=range(len(agg_key)))[0]

    # Final setting of data types before returning the IL input items
    dtypes = {
        **{t: 'float64' for t in ['tiv', 'agg_tiv', 'deductible', 'deductible_min', 'deductible_max', 'limit', 'attachment', 'share',
                                  'deductible1', 'limit1', 'limit2', 'trigger_start', 'trigger_end', 'payout_start', 'payout_end',
                                  'scale1', 'scale2']},
        **{t: 'int32' for t in
           ['agg_id', 'item_id', 'layer_id', 'level_id', 'orig_level_id', 'calcrule_id', 'policytc_id', 'steptriggertype', 'step_id']},
        # **{t: 'uint16' for t in [cond_num]},
        **{t: 'uint8' for t in ['ded_code', 'ded_type', 'lim_code', 'lim_type', 'trigger_type', 'payout_type']}
    }
    il_inputs_df = set_dataframe_column_dtypes(il_inputs_df, dtypes)

    # Assign default values to IL inputs
    il_inputs_df = assign_defaults_to_il_inputs(il_inputs_df)

    # Apply rule to convert type 2 deductibles and limits to TIV shares
    il_inputs_df['deductible'] = np.where(
        il_inputs_df['ded_type'] == DEDUCTIBLE_AND_LIMIT_TYPES['pctiv']['id'],
        il_inputs_df['deductible'] * il_inputs_df['agg_tiv'],
        il_inputs_df['deductible']
    )
    il_inputs_df['limit'] = np.where(
        il_inputs_df['lim_type'] == DEDUCTIBLE_AND_LIMIT_TYPES['pctiv']['id'],
        il_inputs_df['limit'] * il_inputs_df['agg_tiv'],
        il_inputs_df['limit']
    )

    if step_policies_present:
        # Before assigning calc. rule IDs and policy TC IDs, the StepTriggerType
        # should be split into its sub-types in cases where the associated
        # coverages are covered separately
        # For example, StepTriggerType = 5 covers buildings and contents separately
        def assign_sub_step_trigger_type(row):
            try:
                step_trigger_type = STEP_TRIGGER_TYPES[row['steptriggertype']]['sub_step_trigger_types'][
                    row['coverage_type_id']]
                return step_trigger_type
            except KeyError:
                return row['steptriggertype']

        il_inputs_df['StepTriggerType'] = il_inputs_df.apply(
            lambda row: assign_sub_step_trigger_type(row), axis=1
        )

    # Set the calc. rule IDs
    if step_policies_present:
        il_inputs_df.loc[
            ~(il_inputs_df['StepTriggerType'] > 0), 'calcrule_id'
        ] = get_calc_rule_ids(
            il_inputs_df[~(il_inputs_df['StepTriggerType'] > 0)]
        )

        il_inputs_df.loc[
            il_inputs_df['StepTriggerType'] > 0, 'calcrule_id'
        ] = get_step_calc_rule_ids(
            il_inputs_df[il_inputs_df['StepTriggerType'] > 0],
        )
    else:
        il_inputs_df['calcrule_id'] = get_calc_rule_ids(il_inputs_df)

    il_inputs_df['calcrule_id'] = il_inputs_df['calcrule_id'].astype('uint32')

    # Set the policy TC IDs
    if 'cov_agg_id' in il_inputs_df:
        il_inputs_df.loc[
            ~(il_inputs_df['assign_step_calcrule'] > 0), 'policytc_id'
        ] = get_policytc_ids(
            il_inputs_df[~(il_inputs_df['assign_step_calcrule'] > 0)]
        )

        il_inputs_df.loc[
            il_inputs_df['assign_step_calcrule'] > 0, 'policytc_id'
        ] = get_step_policytc_ids(
            il_inputs_df[il_inputs_df['assign_step_calcrule'] > 0],
            offset=il_inputs_df['policytc_id'].max(),
            idx_cols=[acc_num, policy_num, portfolio_num]
        )
    else:
        il_inputs_df['policytc_id'] = get_policytc_ids(il_inputs_df)
    il_inputs_df['policytc_id'] = il_inputs_df['policytc_id'].astype('uint32')

    il_inputs_df = set_dataframe_column_dtypes(il_inputs_df, dtypes)
    return il_inputs_df


@oasis_log
def write_fm_policytc_file(il_inputs_df, fm_policytc_fp, chunksize=100000):
    """
    Writes an FM policy T & C file.

    :param il_inputs_df: IL inputs dataframe
    :type il_inputs_df: pandas.DataFrame

    :param fm_policytc_fp: FM policy TC file path
    :type fm_policytc_fp: str

    :return: FM policy TC file path
    :rtype: str
    """
    try:
        fm_policytc_df = il_inputs_df.loc[il_inputs_df['agg_id'] > 0, ['layer_id', 'level_id', 'agg_id', 'policytc_id']]
        fm_policytc_df.drop_duplicates().to_csv(
            path_or_buf=fm_policytc_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(fm_policytc_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_fm_policytc_file'", e)

    return fm_policytc_fp


@oasis_log
def write_fm_profile_file(il_inputs_df, fm_profile_fp, chunksize=100000):
    """
    Writes an FM profile file.

    :param il_inputs_df: IL inputs dataframe
    :type il_inputs_df: pandas.DataFrame

    :param fm_profile_fp: FM profile file path
    :type fm_profile_fp: str

    :return: FM profile file path
    :rtype: str
    """
    il_inputs_df = il_inputs_df[il_inputs_df['agg_id'] > 0]
    try:
        # Step policies exist
        if 'cov_agg_id' in il_inputs_df:
            fm_profile_df = il_inputs_df[list(set(il_inputs_df.columns).intersection(set(step_profile_cols)))]
            for col in step_profile_cols:
                if col not in fm_profile_df.columns:
                    fm_profile_df[col] = 0

            for non_step_name, step_name in profile_cols_map.items():
                fm_profile_df.loc[
                    ~(il_inputs_df['StepTriggerType'] > 0), step_name
                ] = il_inputs_df.loc[
                    ~(il_inputs_df['StepTriggerType'] > 0),
                    non_step_name
                ]
            fm_profile_df.fillna(0, inplace=True)
            fm_profile_df = fm_profile_df.drop_duplicates()

            # Ensure step_id is of int data type and set default value to 1
            dtypes = {t: 'int64' if t == 'step_id' else 'float64' for t in profile_cols_map.values()}
            fm_profile_df = set_dataframe_column_dtypes(fm_profile_df, dtypes)
            fm_profile_df.loc[fm_profile_df['step_id'] == 0, 'step_id'] = 1

            fm_profile_df.loc[:, step_profile_cols].to_csv(
                path_or_buf=fm_profile_fp,
                encoding='utf-8',
                mode=('w' if os.path.exists(fm_profile_fp) else 'a'),
                chunksize=chunksize,
                index=False
            )
        # No step policies
        else:
            cols = ['policytc_id', 'calcrule_id', 'deductible', 'deductible_min', 'deductible_max', 'attachment', 'limit', 'share']
            fm_profile_df = il_inputs_df.loc[:, cols]

            fm_profile_df.loc[:, cols[2:]] = fm_profile_df.loc[:, cols[2:]].round(7).values

            fm_profile_df.rename(
                columns={
                    'deductible': 'deductible1',
                    'deductible_min': 'deductible2',
                    'deductible_max': 'deductible3',
                    'attachment': 'attachment1',
                    'limit': 'limit1',
                    'share': 'share1'
                },
                inplace=True
            )
            fm_profile_df = fm_profile_df.drop_duplicates()

            fm_profile_df = fm_profile_df.assign(share2=0.0, share3=0.0)

            cols = ['policytc_id', 'calcrule_id', 'deductible1', 'deductible2', 'deductible3', 'attachment1', 'limit1', 'share1', 'share2', 'share3']
            fm_profile_df.loc[:, cols].to_csv(
                path_or_buf=fm_profile_fp,
                encoding='utf-8',
                mode=('w' if os.path.exists(fm_profile_fp) else 'a'),
                chunksize=chunksize,
                index=False
            )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_fm_profile_file'", e)

    return fm_profile_fp


@oasis_log
def write_fm_programme_file(il_inputs_df, fm_programme_fp, chunksize=100000):
    """
    Writes an FM programme file.

    :param il_inputs_df: IL inputs dataframe
    :type il_inputs_df: pandas.DataFrame

    :param fm_programme_fp: FM programme file path
    :type fm_programme_fp: str

    :return: FM programme file path
    :rtype: str
    """
    try:
        il_inputs_df = il_inputs_df[il_inputs_df['to_agg_id'] != 0]
        item_level = il_inputs_df.loc[il_inputs_df['level_id'] == 1, ['gul_input_id', 'level_id', 'agg_id']]
        item_level.rename(columns={'gul_input_id': 'from_agg_id',
                                   'agg_id': 'to_agg_id',
                                   }, inplace=True)
        item_level.drop_duplicates(keep='first', inplace=True)
        fm_programme_df = il_inputs_df[['agg_id', 'level_id', 'to_agg_id']]
        fm_programme_df['level_id'] += 1
        fm_programme_df.rename(columns={'agg_id': 'from_agg_id'}, inplace=True)
        fm_programme_df.drop_duplicates(keep='first', inplace=True)

        fm_programme_df = pd.concat([item_level, fm_programme_df])
        fm_programme_df.to_csv(
            path_or_buf=fm_programme_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(fm_programme_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_fm_programme_file'", e)

    return fm_programme_fp


@oasis_log
def write_fm_xref_file(il_inputs_df, fm_xref_fp, chunksize=100000):
    """
    Writes an FM xref file.

    :param il_inputs_df: IL inputs dataframe
    :type il_inputs_df: pandas.DataFrame

    :param fm_xref_fp: FM xref file path
    :type fm_xref_fp: str

    :return: FM xref file path
    :rtype: str
    """
    try:
        xref_df = get_xref_df(il_inputs_df)
        pd.DataFrame(
            {
                'output': factorize_ndarray(xref_df.loc[:, ['gul_input_id', 'layer_id']].values, col_idxs=range(2))[0],
                'agg_id': xref_df['gul_input_id'],
                'layer_id': xref_df['layer_id']
            }
        ).to_csv(
            path_or_buf=fm_xref_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(fm_xref_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_fm_xref_file'", e)

    return fm_xref_fp


@oasis_log
def write_il_input_files(
        il_inputs_df,
        target_dir,
        oasis_files_prefixes=copy.deepcopy(OASIS_FILES_PREFIXES['il']),
        chunksize=(2 * 10 ** 5)
):
    """
    Writes standard Oasis IL input files to a target directory using a
    pre-generated dataframe of IL inputs dataframe. The files written are
    ::

        fm_policytc.csv
        fm_profile.csv
        fm_programme.csv
        fm_xref.csv

    :param il_inputs_df: IL inputs dataframe
    :type exposure_df: pandas.DataFrame

    :param target_dir: Target directory in which to write the files
    :type target_dir: str

    :param oasis_files_prefixes: Oasis IL input file name prefixes
    :param oasis_files_prefixes: dict

    :param chunksize: The chunk size to use when writing out the
                      input files
    :type chunksize: int

    :return: IL input files dict
    :rtype: dict
    """
    # Clean the target directory path
    target_dir = as_path(target_dir, 'Target IL input files directory', is_dir=True, preexists=False)

    # Set chunk size for writing the CSV files - default is the minimum of 100K
    # or the IL inputs frame size
    chunksize = chunksize or min(2 * 10 ** 5, len(il_inputs_df))

    # A dict of IL input file names and file paths
    il_input_files = {
        fn: os.path.join(target_dir, '{}.csv'.format(oasis_files_prefixes[fn])) for fn in oasis_files_prefixes
    }

    this_module = sys.modules[__name__]
    # Write the files serially
    for fn in il_input_files:
        getattr(this_module, 'write_{}_file'.format(fn))(il_inputs_df.copy(deep=True), il_input_files[fn], chunksize)

    return il_input_files
